# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Internal utility classes and functions used by Tanjun."""
from __future__ import annotations

__all__: list[str] = []

import asyncio
import copy as copy_
import enum
import functools
import itertools
import logging
import operator
import sys
import types
import typing
from collections import abc as collections

import hikari

from .. import errors
from .vendor import inspect

if typing.TYPE_CHECKING:
    import typing_extensions

    from .. import abc as tanjun

    _T = typing.TypeVar("_T")
    _P = typing_extensions.ParamSpec("_P")

    _ContextT = typing.TypeVar("_ContextT", bound=tanjun.Context)
    _CoroT = collections.Coroutine[typing.Any, typing.Any, _T]
    _TreeT = dict[
        typing.Union[str, "_IndexKeys"],
        typing.Union["_TreeT", list[tuple[list[str], tanjun.MessageCommand[typing.Any]]]],
    ]


_KeyT = typing.TypeVar("_KeyT")
_OtherT = typing.TypeVar("_OtherT")

_LOGGER = logging.getLogger("hikari.tanjun")

if sys.version_info >= (3, 10):
    UnionTypes = frozenset((typing.Union, types.UnionType))

else:
    UnionTypes = frozenset((typing.Union,))


class _DefaultEnum(enum.Enum):
    VALUE = object()


DEFAULT = _DefaultEnum.VALUE
"""Internal singleton used to signify when a value wasn't provided."""

Default = typing.Literal[_DefaultEnum.VALUE]
"""The type of `DEFAULT`."""


async def _execute_check(ctx: _ContextT, callback: tanjun.CheckSig[_ContextT], /) -> bool:
    if result := await ctx.call_with_async_di(callback, ctx):
        return result

    raise errors.FailedCheck


async def gather_checks(ctx: _ContextT, checks: collections.Iterable[tanjun.CheckSig[_ContextT]], /) -> bool:
    """Gather a collection of checks.

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The context to check.
    checks : collections.abc.Iterable[tanjun.abc.CheckSig]
        An iterable of injectable checks.

    Returns
    -------
    bool
        Whether all the checks passed or not.
    """
    try:
        await asyncio.gather(*(_execute_check(ctx, check) for check in checks))

    except errors.FailedCheck:
        return False

    else:
        # InjectableCheck will raise FailedCheck if a false is received so if
        # we get this far then it's True.
        return True


_EMPTY_BUFFER: dict[typing.Any, typing.Any] = {}


class CastedView(collections.Mapping[_KeyT, _OtherT]):
    """Utility class for exposing an immutable casted view of a dict."""

    __slots__ = ("_buffer", "_cast", "_raw_data")

    def __init__(self, raw_data: dict[_KeyT, _T], cast: collections.Callable[[_T], _OtherT], /) -> None:
        self._buffer: dict[_KeyT, _OtherT] = {} if raw_data else _EMPTY_BUFFER
        self._cast = cast
        self._raw_data = raw_data

    def __getitem__(self, key: _KeyT, /) -> _OtherT:
        try:
            return self._buffer[key]

        except KeyError:
            pass

        entry = self._raw_data[key]
        result = self._cast(entry)
        self._buffer[key] = result
        return result

    def __iter__(self) -> collections.Iterator[_KeyT]:
        return iter(self._raw_data)

    def __len__(self) -> int:
        return len(self._raw_data)


_KEYWORD_TYPES = {inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD}


def get_kwargs(callback: collections.Callable[..., typing.Any], /) -> typing.Union[list[str], None]:
    """Get a list of the keyword argument names for a callback.

    Parameters
    ----------
    callback
        The callback to get the keyword argument names of.

    Returns
    -------
    list[str] | None
        A list of the keyword argument names for this callback or [None][]
        if this argument takes `**kwargs`.
    """
    names: list[str] = []

    try:
        signature = inspect.Signature.from_callable(callback)

    except ValueError:
        # When "no signature [is] found" for a callback/type, we just don't
        # know what parameters it has so we have to assume var keyword.
        return None

    for parameter in signature.parameters.values():
        if parameter.kind is parameter.VAR_KEYWORD:
            return None

        if parameter.kind in _KEYWORD_TYPES:
            names.append(parameter.name)

    return names


_POSITIONAL_TYPES = {
    inspect.Parameter.POSITIONAL_ONLY,
    inspect.Parameter.POSITIONAL_OR_KEYWORD,
    inspect.Parameter.VAR_POSITIONAL,
}


def _snoop_types(type_: typing.Any, /) -> collections.Iterator[typing.Any]:
    origin = typing.get_origin(type_)
    if origin in UnionTypes:
        yield from itertools.chain.from_iterable(map(_snoop_types, typing.get_args(type_)))

    elif origin is typing.Annotated:
        yield from _snoop_types(typing.get_args(type_)[0])

    else:
        yield type_


def infer_listener_types(
    callback: collections.Callable[..., collections.Coroutine[typing.Any, typing.Any, None]], /
) -> collections.Sequence[type[hikari.Event]]:
    """Infer the event type(s) of an event listener callback from its annotations.

    Arguments
    ---------
    callback
        The callback to infer the listener types for.

    Returns
    -------
    collections.Sequence[type[hikari.Event]]
        Sequence of the listener types for this callback.

    Raises
    ------
    ValueError
        If the callback's first argument isn't positional or doesn't have
        a type hint.
    TypeError
        If the callback's first positional argument doesn't have any
        [hiari.events.base_events.Event][] subclasses in it.
    """
    try:
        signature = inspect.Signature.from_callable(callback, eval_str=True)
    except ValueError:  # Callback has no signature
        raise ValueError("Missing event type") from None

    parameter = next(iter(signature.parameters.values()), None)
    if not parameter or parameter.kind not in _POSITIONAL_TYPES:
        raise ValueError("Missing positional event argument") from None

    if parameter.annotation is parameter.empty:
        raise ValueError("Missing event argument annotation") from None

    event_types: list[type[hikari.Event]] = []

    for type_ in _snoop_types(parameter.annotation):
        try:
            if issubclass(type_, hikari.Event):
                event_types.append(type_)

        except TypeError:
            pass

    if not event_types:
        raise TypeError(f"No valid event types found in the signature of {callback}") from None

    return event_types


def log_task_exc(
    message: str, /
) -> collections.Callable[[collections.Callable[_P, collections.Awaitable[_T]]], collections.Callable[_P, _CoroT[_T]]]:
    """Log the exception when a task raises instead of leaving it up to the gods."""

    def decorator(
        callback: collections.Callable[_P, collections.Awaitable[_T]], /
    ) -> collections.Callable[_P, _CoroT[_T]]:
        @functools.wraps(callback)
        async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            try:
                return await callback(*args, **kwargs)

            except Exception as exc:
                _LOGGER.exception(message, exc_info=exc)
                raise

        return wrapper

    return decorator


class _WrappedProto(typing.Protocol):
    wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]]


def _has_wrapped(value: typing.Any, /) -> typing_extensions.TypeGuard[_WrappedProto]:
    try:
        value.wrapped_command

    except AttributeError:
        return False

    return True


def collect_wrapped(command: tanjun.ExecutableCommand[typing.Any], /) -> list[tanjun.ExecutableCommand[typing.Any]]:
    """Collect all the commands a command object has wrapped in decorator calls.

    Parameters
    ----------
    command
        The top-level command object.

    Returns
    -------
    list[tanjun.abc.ExecutableCommand[typing.Any]]
        A list of the wrapped commands.
    """
    results: list[tanjun.ExecutableCommand[typing.Any]] = []
    wrapped = command.wrapped_command if _has_wrapped(command) else None

    while wrapped:
        results.append(wrapped)
        wrapped = wrapped.wrapped_command if _has_wrapped(wrapped) else None

    return results


_OptionT = typing.TypeVar("_OptionT", bound=hikari.CommandInteractionOption)
SUB_COMMAND_OPTION_TYPES: typing.Final[frozenset[hikari.OptionType]] = frozenset(
    [hikari.OptionType.SUB_COMMAND, hikari.OptionType.SUB_COMMAND_GROUP]
)


def flatten_options(
    name: str, options: typing.Optional[collections.Sequence[_OptionT]], /
) -> tuple[str, collections.Sequence[_OptionT]]:
    """Flatten the options of a slash/autocomplete interaction.

    Parameters
    ----------
    options
        The options to flatten.

    Returns
    -------
    tuple[str, collections.abc.Sequence[_OptionT]]
        The full triggering command name and a sequence of the actual command options.
    """
    while options and (first_option := options[0]).type in SUB_COMMAND_OPTION_TYPES:
        name = f"{name} {first_option.name}"
        options = typing.cast("collections.Sequence[_OptionT]", first_option.options)

    return name, options or ()


CHANNEL_TYPES: dict[type[hikari.PartialChannel], set[hikari.ChannelType]] = {
    hikari.GuildTextChannel: {hikari.ChannelType.GUILD_TEXT},
    hikari.DMChannel: {hikari.ChannelType.DM},
    hikari.GuildVoiceChannel: {hikari.ChannelType.GUILD_VOICE},
    hikari.GroupDMChannel: {hikari.ChannelType.GROUP_DM},
    hikari.GuildCategory: {hikari.ChannelType.GUILD_CATEGORY},
    hikari.GuildNewsChannel: {hikari.ChannelType.GUILD_NEWS},
    hikari.GuildStageChannel: {hikari.ChannelType.GUILD_STAGE},
    hikari.GuildNewsThread: {hikari.ChannelType.GUILD_NEWS_THREAD},
    hikari.GuildPublicThread: {hikari.ChannelType.GUILD_PUBLIC_THREAD},
    hikari.GuildPrivateThread: {hikari.ChannelType.GUILD_PRIVATE_THREAD},
    hikari.GuildForumChannel: {hikari.ChannelType.GUILD_FORUM},
}
"""Mapping of hikari channel classes to the raw channel types which are compatible for it."""


for _channel_cls, _types in CHANNEL_TYPES.copy().items():
    for _mro_type in _channel_cls.mro():
        if isinstance(_mro_type, type) and issubclass(_mro_type, hikari.PartialChannel):
            try:
                CHANNEL_TYPES[_mro_type].update(_types)
            except KeyError:
                CHANNEL_TYPES[_mro_type] = _types.copy()

# This isn't a base class but it should still act like an indicator for any channel type.
CHANNEL_TYPES[hikari.InteractionChannel] = CHANNEL_TYPES[hikari.PartialChannel]


def parse_channel_types(*channel_types: typing.Union[type[hikari.PartialChannel], int]) -> list[hikari.ChannelType]:
    """Parse a channel types collection to a list of channel type integers."""
    types_iter = itertools.chain.from_iterable(
        (hikari.ChannelType(type_),) if isinstance(type_, int) else CHANNEL_TYPES[type_] for type_ in channel_types
    )

    try:
        return list(dict.fromkeys(types_iter))

    except KeyError as exc:
        raise ValueError(f"Unknown channel type {exc.args[0]}") from exc


_CHANNEL_TYPE_REPS: dict[hikari.ChannelType, str] = {
    hikari.ChannelType.GUILD_TEXT: "Text",
    hikari.ChannelType.DM: "DM",
    hikari.ChannelType.GUILD_VOICE: "Voice",
    hikari.ChannelType.GROUP_DM: "Group DM",
    hikari.ChannelType.GUILD_CATEGORY: "Category",
    hikari.ChannelType.GUILD_NEWS: "News",
    hikari.ChannelType.GUILD_STAGE: "Stage",
    hikari.ChannelType.GUILD_NEWS_THREAD: "News Thread",
    hikari.ChannelType.GUILD_PUBLIC_THREAD: "Public Thread",
    hikari.ChannelType.GUILD_PRIVATE_THREAD: "Private Thread",
    hikari.ChannelType.GUILD_FORUM: "Forum",
}
_UNKNOWN_CHANNEL_REPR = "Unknown"


def repr_channel(channel_type: hikari.ChannelType, /) -> str:
    """Get a text repr of a channel type."""
    return _CHANNEL_TYPE_REPS.get(channel_type, _UNKNOWN_CHANNEL_REPR)


def cmp_command(
    cmd: typing.Union[hikari.PartialCommand, hikari.api.CommandBuilder],
    other: typing.Union[hikari.PartialCommand, hikari.api.CommandBuilder, None],
    /,
) -> bool:
    """Compare application command objects and command builders."""
    if not other or other.type != cmd.type:
        return False

    dm_enabled = True if cmd.is_dm_enabled is hikari.UNDEFINED else cmd.is_dm_enabled
    other_dm_enabled = True if other.is_dm_enabled is hikari.UNDEFINED else other.is_dm_enabled
    default_perms = cmd.default_member_permissions or hikari.Permissions.NONE
    other_default_perms = other.default_member_permissions or hikari.Permissions.NONE

    if dm_enabled is not other_dm_enabled or default_perms != other_default_perms:
        return False

    # name doesn't need to be checked as `builder` will be `None` if that didn't match.
    if cmd.name_localizations != other.name_localizations:
        return False

    if isinstance(cmd, (hikari.SlashCommand, hikari.api.SlashCommandBuilder)):
        assert isinstance(other, (hikari.SlashCommand, hikari.api.SlashCommandBuilder))
        if cmd.description != other.description or cmd.description_localizations != other.description_localizations:
            return False

        opts = cmd.options or ()
        other_opts = other.options or ()
        return len(opts) == len(other_opts) and all(itertools.starmap(operator.eq, zip(opts, other_opts)))

    return True


def cmp_all_commands(
    commands: collections.Collection[typing.Union[hikari.PartialCommand, hikari.api.CommandBuilder]],
    other: collections.Mapping[
        tuple[hikari.CommandType, str], typing.Union[hikari.PartialCommand, hikari.api.CommandBuilder]
    ],
    /,
) -> bool:
    """Compare two sets of command objects/builders."""
    return len(commands) == len(other) and all(cmp_command(c, other.get((c.type, c.name))) for c in commands)


class _IndexKeys(enum.Enum):
    COMMANDS = enum.auto()
    PARENT = enum.auto()


class MessageCommandIndex:
    """A searchable message command index."""

    __slots__ = ("commands", "is_strict", "names_to_commands", "search_tree")

    def __init__(
        self,
        strict: bool,
        /,
        *,
        commands: typing.Optional[list[tanjun.MessageCommand[typing.Any]]] = None,
        names_to_commands: typing.Optional[dict[str, tuple[str, tanjun.MessageCommand[typing.Any]]]] = None,
        search_tree: typing.Optional[_TreeT] = None,
    ) -> None:
        """Initialise a message command index.

        Parameters
        ----------
        strict
            Whether the index should be strict about command names.

            Command names must be (case-insensitively) unique in a strict index and
            must not contain spaces.
        """
        self.commands = commands or []
        self.is_strict = strict
        self.names_to_commands = names_to_commands or {}
        self.search_tree = search_tree or {}

    def add(self, command: tanjun.MessageCommand[typing.Any], /) -> bool:
        """Add a command to the index.

        Parameters
        ----------
        command
            The command to add.

        Returns
        -------
        bool
            Whether the command was added.

            If this is [False][] then the command was already in the index.

        Raises
        ------
        ValueError
            If the command name is invalid or already in the index.
        """
        if command in self.commands:
            return False

        if self.is_strict:
            if any(" " in name for name in command.names):
                raise ValueError("Command name cannot contain spaces in a strict collection")

            names = list(filter(None, command.names))
            insensitive_names = [name.casefold() for name in command.names]
            if name_conflicts := self.names_to_commands.keys() & insensitive_names:
                raise ValueError(
                    "Sub-command names must be (case-insensitively) unique in a strict collection. "
                    "The following conflicts were found " + ", ".join(name_conflicts)
                )

            # Case insensitive keys are used here as a subsequent check against the original
            # name can be used for case-sensitive lookup.
            self.names_to_commands.update((key, (name, command)) for key, name in zip(insensitive_names, names))

        else:  # strict indexes avoid using the search tree all together.
            # This needs to be explicitly typed for MyPy.
            node: typing.Union[_TreeT, list[tuple[list[str], tanjun.MessageCommand[typing.Any]]]]
            for name in filter(None, command.names):
                node = self.search_tree
                # The search tree is kept case-insensitive as a check against the actual name
                # can be used to ensure case-sensitivity.
                for chars in name.casefold().split(" "):
                    try:
                        node = node[chars]

                    except KeyError:
                        new_node: _TreeT = {_IndexKeys.PARENT: node}
                        node[chars] = node = new_node

                    assert isinstance(node, dict)

                # A case-preserved variant of the name is stored alongside the command
                # for case-sensitive lookup
                # This is split into a list of words to avoid mult-spaces failing lookup.
                name_parts = name.split(" ")
                try:
                    commands = node[_IndexKeys.COMMANDS]
                    assert isinstance(commands, list)
                    commands.append((name_parts, command))
                except KeyError:
                    node[_IndexKeys.COMMANDS] = [(name_parts, command)]

        self.commands.append(command)
        return True

    def copy(self, *, parent: typing.Optional[tanjun.MessageCommandGroup[typing.Any]] = None) -> MessageCommandIndex:
        """In-place copy the index and its contained commands.

        Parameters
        ----------
        parent
            The parent message command group of the copied index.

        Returns
        -------
        MessageCommandIndex
            The copied index.
        """
        commands = {command: command.copy(parent=parent) for command in self.commands}
        memo = {id(command): new_command for command, new_command in commands.items()}
        return MessageCommandIndex(
            self.is_strict,
            commands=list(commands.values()),
            names_to_commands={
                key: (name, commands[command]) for key, (name, command) in self.names_to_commands.items()
            },
            search_tree=copy_.deepcopy(self.search_tree, memo),
        )

    def find(
        self, content: str, case_sensitive: bool, /
    ) -> collections.Iterator[tuple[str, tanjun.MessageCommand[typing.Any]]]:
        """Find commands in the index.

        Parameters
        ----------
        content
            The content to search for.
        case_sensitive
            Whether the search should be case-sensitive.

        Yields
        ------
        tuple[str, tanjun.abc.MessageCommand[typing.Any]]
            A tuple of the matching name and command.
        """
        if self.is_strict:
            name = content.split(" ", 1)[0]
            # A case-insensitive key is used to allow for both the case-sensitive and case-insensitive
            # cases to be covered.
            command = self.names_to_commands.get(name.casefold())
            if command and (not case_sensitive or command[0] == name):
                yield name, command[1]

            # strict indexes avoid using the search tree all together.
            return

        # This needs to be explicitly typed for MyPy.
        node: typing.Union[_TreeT, list[tuple[list[str], tanjun.MessageCommand[typing.Any]]]]
        node = self.search_tree
        segments: list[tuple[int, list[tuple[list[str], tanjun.MessageCommand[typing.Any]]]]] = []
        split = content.split(" ")
        for index, chars in enumerate(split):
            try:
                node = node[chars.casefold()]

            except KeyError:
                break

            else:
                assert isinstance(node, dict)
                if entries := node.get(_IndexKeys.COMMANDS):
                    assert isinstance(entries, list)
                    segments.append((index, entries))

        for index, segment in reversed(segments):
            name_parts = split[: index + 1]
            name = " ".join(name_parts)
            if case_sensitive:
                yield from ((name, c) for n, c in segment if n == name_parts)

            else:
                yield from ((name, c) for _, c in segment)

    def remove(self, command: tanjun.MessageCommand[typing.Any], /) -> None:
        """Remove a command from the index.

        Parameters
        ----------
        command
            The command to remove.

        Raises
        ------
        ValueError
            If the command is not in the index.
        """
        self.commands.remove(command)

        if self.is_strict:
            for name in map(str.casefold, filter(None, command.names)):
                if (entry := self.names_to_commands.get(name)) and entry[1] == command:
                    del self.names_to_commands[name]

            # strict indexes avoid using the search tree all together.
            return

        # This needs to be explicitly typed for MyPy.
        node: typing.Union[_TreeT, list[tuple[list[str], tanjun.MessageCommand[typing.Any]]]]
        for name in filter(None, command.names):
            nodes: list[tuple[str, _TreeT]] = []
            node = self.search_tree
            for chars in name.casefold().split(" "):
                try:
                    node = node[chars]

                except KeyError:
                    # The command is not in the index and we want to skip the "else" statement.
                    break

                else:
                    assert isinstance(node, dict)
                    nodes.append((chars, node))

            else:
                # If it didn't break out of the for chars loop then the command is in here.
                name_parts = name.split(" ")
                entries = node[_IndexKeys.COMMANDS]
                assert isinstance(entries, list)
                entries.remove((name_parts, command))  # Remove the command from the last node.
                if not node[_IndexKeys.COMMANDS]:
                    del node[_IndexKeys.COMMANDS]

                # Otherwise, we need to remove the node from the tree.
                for chars, node in reversed(nodes):
                    if len(node) > 1:
                        # If the node is not empty then we're done.
                        continue

                    parent = node.get(_IndexKeys.PARENT)
                    if not parent:
                        break

                    assert isinstance(parent, dict)
                    del parent[chars]



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Dependency used for managing owner checks."""
from __future__ import annotations

__all__: list[str] = ["AbstractOwners", "Owners"]

import abc
import asyncio
import datetime
import logging
import time
import typing

import hikari

from . import async_cache

if typing.TYPE_CHECKING:
    from collections import abc as collections

    from .. import abc as tanjun


_T = typing.TypeVar("_T")
_LOGGER: typing.Final[logging.Logger] = logging.getLogger("hikari.tanjun")


class AbstractOwners(abc.ABC):
    """Interface used to check if a user is deemed to be the bot's "owner"."""

    __slots__ = ()

    @abc.abstractmethod
    async def check_ownership(self, client: tanjun.Client, user: hikari.User, /) -> bool:
        """Check whether this object is owned by the given object.

        Parameters
        ----------
        client
            The Tanjun client this check is being called by.
        user
            The user to check ownership for.

        Returns
        -------
        bool
            Whether the bot is owned by the provided user.
        """


class _CachedValue(typing.Generic[_T]):
    __slots__ = ("_expire_after", "_last_called", "_lock", "_result")

    def __init__(self, *, expire_after: typing.Optional[float]) -> None:
        self._expire_after = expire_after
        self._last_called: typing.Optional[float] = None
        self._lock: typing.Optional[asyncio.Lock] = None
        self._result: typing.Optional[_T] = None

    @property
    def _has_expired(self) -> bool:
        return self._expire_after is not None and (
            not self._last_called or self._expire_after <= (time.monotonic() - self._last_called)
        )

    async def acquire(self, callback: collections.Callable[[], collections.Awaitable[_T]], /) -> _T:
        if self._result is not None and not self._has_expired:
            return self._result

        if self._lock is None:
            self._lock = asyncio.Lock()

        async with self._lock:
            if self._result is not None and not self._has_expired:
                return self._result

            self._result = await callback()
            self._last_called = time.monotonic()
            # This is set to None afterwards to ensure that it isn't persisted between loops.
            self._lock = None
            return self._result


_ApplicationCacheT = async_cache.SingleStoreCache[hikari.Application]


class Owners(AbstractOwners):
    """Default implementation of the owner check interface.

    !!! warning
        `fallback_to_application` is only possible when the REST client
        is bound to a Bot token or if a type dependency is registered for
        `tanjun.dependencies.SingleStoreCache[hikari.Application]`.
    """

    __slots__ = ("_fallback_to_application", "_owner_ids", "_value")

    def __init__(
        self,
        *,
        expire_after: typing.Union[datetime.timedelta, int, float] = datetime.timedelta(minutes=5),
        fallback_to_application: bool = True,
        owners: typing.Optional[hikari.SnowflakeishSequence[hikari.User]] = None,
    ) -> None:
        """Initiate a new owner check dependency.

        Parameters
        ----------
        expire_after
            The amount of time to cache application owner data for in seconds.

            This is only applicable if `rest` is also passed.
        fallback_to_application
            Whether this check should fallback to checking the application's owners
            if the user isn't in `owners`.

            This only works when the bot's rest client is bound to a Bot token or
            if `tanjun.dependencies.SingleStoreCache[hikari.Application]` is available.
        owners
            Sequence of objects and IDs of the users that are allowed to use the
            bot's owners-only commands.
        """
        if isinstance(expire_after, datetime.timedelta):
            expire_after = expire_after.total_seconds()
        else:
            expire_after = float(expire_after)

        if expire_after <= 0:
            raise ValueError("Expire after must be greater than 0 seconds")

        self._fallback_to_application = fallback_to_application
        self._owner_ids = {hikari.Snowflake(id_) for id_ in owners} if owners else set[hikari.Snowflake]()
        self._value = _CachedValue[hikari.Application](expire_after=expire_after)

    async def check_ownership(self, client: tanjun.Client, user: hikari.User, /) -> bool:
        if user.id in self._owner_ids:
            return True

        if not self._fallback_to_application:
            return False

        application_cache = client.get_type_dependency(_ApplicationCacheT)
        if application_cache and (application := await application_cache.get(default=None)):
            return user.id in application.team.members if application.team else user.id == application.owner.id

        if client.rest.token_type is not hikari.TokenType.BOT:
            _LOGGER.warning(
                "Owner checks cannot fallback to application owners when bound to an OAuth2 "
                "client credentials token and may always fail unless bound to a Bot token."
            )
            return False

        application = await self._value.acquire(client.rest.fetch_application)
        return user.id in application.team.members if application.team else user.id == application.owner.id



# See ./inspect.pyi.LICENSE for license information.

# Vendored from https://github.com/python/typeshed/blob/0e185f40877a614c5ddd87815707cf8153296872/stdlib/inspect.pyi
# with modifications which freeze it to 3.10's typeshed.
import enum
import sys
import types
from collections import OrderedDict
from collections.abc import Awaitable, Callable, Generator, Mapping, Sequence, Set as AbstractSet
from types import (
    AsyncGeneratorType,
    BuiltinFunctionType,
    BuiltinMethodType,
    CodeType,
    CoroutineType,
    FrameType,
    FunctionType,
    GeneratorType,
    GetSetDescriptorType,
    LambdaType,
    MethodType,
    ModuleType,
    TracebackType,
)

if sys.version_info >= (3, 7):
    from types import ClassMethodDescriptorType, WrapperDescriptorType, MemberDescriptorType, MethodDescriptorType

from typing import Any, ClassVar, Coroutine, NamedTuple, Protocol, TypeVar, Union
from typing_extensions import Literal, ParamSpec, Self, TypeGuard

_P = ParamSpec("_P")
_T_cont = TypeVar("_T_cont", contravariant=True)
_V_cont = TypeVar("_V_cont", contravariant=True)

#
# Types and members
#
class EndOfBlock(Exception): ...

class BlockFinder:
    indent: int
    islambda: bool
    started: bool
    passline: bool
    indecorator: bool
    decoratorhasargs: bool
    last: int
    def tokeneater(self, type: int, token: str, srowcol: tuple[int, int], erowcol: tuple[int, int], line: str) -> None: ...

CO_OPTIMIZED: Literal[1]
CO_NEWLOCALS: Literal[2]
CO_VARARGS: Literal[4]
CO_VARKEYWORDS: Literal[8]
CO_NESTED: Literal[16]
CO_GENERATOR: Literal[32]
CO_NOFREE: Literal[64]
CO_COROUTINE: Literal[128]
CO_ITERABLE_COROUTINE: Literal[256]
CO_ASYNC_GENERATOR: Literal[512]
TPFLAGS_IS_ABSTRACT: Literal[1048576]

modulesbyfile: dict[str, Any]

def getmembers(object: object, predicate: Callable[[Any], bool] | None = ...) -> list[tuple[str, Any]]: ...
def getmodulename(path: str) -> str | None: ...
def ismodule(object: object) -> TypeGuard[ModuleType]: ...
def isclass(object: object) -> TypeGuard[type[Any]]: ...
def ismethod(object: object) -> TypeGuard[MethodType]: ...
def isfunction(object: object) -> TypeGuard[FunctionType]: ...

if sys.version_info >= (3, 8):
    def isgeneratorfunction(obj: object) -> bool: ...
    def iscoroutinefunction(obj: object) -> bool: ...

else:
    def isgeneratorfunction(object: object) -> bool: ...
    def iscoroutinefunction(object: object) -> bool: ...

def isgenerator(object: object) -> TypeGuard[GeneratorType[Any, Any, Any]]: ...
def iscoroutine(object: object) -> TypeGuard[CoroutineType[Any, Any, Any]]: ...
def isawaitable(object: object) -> TypeGuard[Awaitable[Any]]: ...

if sys.version_info >= (3, 8):
    def isasyncgenfunction(obj: object) -> bool: ...

else:
    def isasyncgenfunction(object: object) -> bool: ...

class _SupportsSet(Protocol[_T_cont, _V_cont]):
    def __set__(self, __instance: _T_cont, __value: _V_cont) -> None: ...

class _SupportsDelete(Protocol[_T_cont]):
    def __delete__(self, __instance: _T_cont) -> None: ...

def isasyncgen(object: object) -> TypeGuard[AsyncGeneratorType[Any, Any]]: ...
def istraceback(object: object) -> TypeGuard[TracebackType]: ...
def isframe(object: object) -> TypeGuard[FrameType]: ...
def iscode(object: object) -> TypeGuard[CodeType]: ...
def isbuiltin(object: object) -> TypeGuard[BuiltinFunctionType]: ...

if sys.version_info >= (3, 7):
    def isroutine(
        object: object,
    ) -> TypeGuard[
        FunctionType
        | LambdaType
        | MethodType
        | BuiltinFunctionType
        | BuiltinMethodType
        | WrapperDescriptorType
        | MethodDescriptorType
        | ClassMethodDescriptorType
    ]: ...
    def ismethoddescriptor(object: object) -> TypeGuard[MethodDescriptorType]: ...
    def ismemberdescriptor(object: object) -> TypeGuard[MemberDescriptorType]: ...

else:
    def isroutine(
        object: object,
    ) -> TypeGuard[FunctionType | LambdaType | MethodType | BuiltinFunctionType | BuiltinMethodType]: ...
    def ismethoddescriptor(object: object) -> bool: ...
    def ismemberdescriptor(object: object) -> bool: ...

def isabstract(object: object) -> bool: ...
def isgetsetdescriptor(object: object) -> TypeGuard[GetSetDescriptorType]: ...
def isdatadescriptor(object: object) -> TypeGuard[_SupportsSet[Any, Any] | _SupportsDelete[Any]]: ...

#
# Retrieving source code
#
_SourceObjectType = Union[ModuleType, type[Any], MethodType, FunctionType, TracebackType, FrameType, CodeType, Callable[..., Any]]

def findsource(object: _SourceObjectType) -> tuple[list[str], int]: ...
def getabsfile(object: _SourceObjectType, _filename: str | None = ...) -> str: ...
def getblock(lines: Sequence[str]) -> Sequence[str]: ...
def getdoc(object: object) -> str | None: ...
def getcomments(object: object) -> str | None: ...
def getfile(object: _SourceObjectType) -> str: ...
def getmodule(object: object, _filename: str | None = ...) -> ModuleType | None: ...
def getsourcefile(object: _SourceObjectType) -> str | None: ...
def getsourcelines(object: _SourceObjectType) -> tuple[list[str], int]: ...
def getsource(object: _SourceObjectType) -> str: ...
def cleandoc(doc: str) -> str: ...
def indentsize(line: str) -> int: ...

#
# Introspecting callables with the Signature object
#
def signature(
    obj: Callable[..., Any],
    *,
    follow_wrapped: bool = ...,
    globals: Mapping[str, Any] | None = ...,
    locals: Mapping[str, Any] | None = ...,
    eval_str: bool = ...,
) -> Signature: ...

class _void: ...
class _empty: ...

class Signature:
    def __init__(
        self, parameters: Sequence[Parameter] | None = ..., *, return_annotation: Any = ..., __validate_parameters__: bool = ...
    ) -> None: ...
    empty = _empty
    @property
    def parameters(self) -> types.MappingProxyType[str, Parameter]: ...
    @property
    def return_annotation(self) -> Any: ...
    def bind(self, *args: Any, **kwargs: Any) -> BoundArguments: ...
    def bind_partial(self, *args: Any, **kwargs: Any) -> BoundArguments: ...
    def replace(
        self, *, parameters: Sequence[Parameter] | type[_void] | None = ..., return_annotation: Any = ...
    ) -> Self: ...
    @classmethod
    def from_callable(
        cls,
        obj: Callable[..., Any],
        *,
        follow_wrapped: bool = ...,
        globals: Mapping[str, Any] | None = ...,
        locals: Mapping[str, Any] | None = ...,
        eval_str: bool = ...,
    ) -> Self: ...

def get_annotations(
    obj: Callable[..., Any] | type[Any] | ModuleType,
    *,
    globals: Mapping[str, Any] | None = ...,
    locals: Mapping[str, Any] | None = ...,
    eval_str: bool = ...,
) -> dict[str, Any]: ...

# The name is the same as the enum's name in CPython
class _ParameterKind(enum.IntEnum):
    POSITIONAL_ONLY: int
    POSITIONAL_OR_KEYWORD: int
    VAR_POSITIONAL: int
    KEYWORD_ONLY: int
    VAR_KEYWORD: int

    if sys.version_info >= (3, 8):
        @property
        def description(self) -> str: ...

class Parameter:
    def __init__(self, name: str, kind: _ParameterKind, *, default: Any = ..., annotation: Any = ...) -> None: ...
    empty = _empty

    POSITIONAL_ONLY: ClassVar[Literal[_ParameterKind.POSITIONAL_ONLY]]
    POSITIONAL_OR_KEYWORD: ClassVar[Literal[_ParameterKind.POSITIONAL_OR_KEYWORD]]
    VAR_POSITIONAL: ClassVar[Literal[_ParameterKind.VAR_POSITIONAL]]
    KEYWORD_ONLY: ClassVar[Literal[_ParameterKind.KEYWORD_ONLY]]
    VAR_KEYWORD: ClassVar[Literal[_ParameterKind.VAR_KEYWORD]]
    @property
    def name(self) -> str: ...
    @property
    def default(self) -> Any: ...
    @property
    def kind(self) -> _ParameterKind: ...
    @property
    def annotation(self) -> Any: ...
    def replace(
        self,
        *,
        name: str | type[_void] = ...,
        kind: _ParameterKind | type[_void] = ...,
        default: Any = ...,
        annotation: Any = ...,
    ) -> Self: ...

class BoundArguments:
    arguments: OrderedDict[str, Any]
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    signature: Signature
    def __init__(self, signature: Signature, arguments: OrderedDict[str, Any]) -> None: ...
    def apply_defaults(self) -> None: ...

#
# Classes and functions
#

# TODO: The actual return type should be list[_ClassTreeItem] but mypy doesn't
# seem to be supporting this at the moment:
# _ClassTreeItem = list[_ClassTreeItem] | Tuple[type, Tuple[type, ...]]
def getclasstree(classes: list[type], unique: bool = ...) -> list[Any]: ...
def walktree(classes: list[type], children: dict[type[Any], list[type]], parent: type[Any] | None) -> list[Any]: ...

class Arguments(NamedTuple):
    args: list[str]
    varargs: str | None
    varkw: str | None

def getargs(co: CodeType) -> Arguments: ...

class FullArgSpec(NamedTuple):
    args: list[str]
    varargs: str | None
    varkw: str | None
    defaults: tuple[Any, ...] | None
    kwonlyargs: list[str]
    kwonlydefaults: dict[str, Any] | None
    annotations: dict[str, Any]

def getfullargspec(func: object) -> FullArgSpec: ...

class ArgInfo(NamedTuple):
    args: list[str]
    varargs: str | None
    keywords: str | None
    locals: dict[str, Any]

def getargvalues(frame: FrameType) -> ArgInfo: ...
def formatannotation(annotation: object, base_module: str | None = ...) -> str: ...
def formatannotationrelativeto(object: object) -> Callable[[object], str]: ...

def formatargvalues(
    args: list[str],
    varargs: str | None,
    varkw: str | None,
    locals: dict[str, Any] | None,
    formatarg: Callable[[str], str] | None = ...,
    formatvarargs: Callable[[str], str] | None = ...,
    formatvarkw: Callable[[str], str] | None = ...,
    formatvalue: Callable[[Any], str] | None = ...,
) -> str: ...
def getmro(cls: type) -> tuple[type, ...]: ...
def getcallargs(__func: Callable[_P, Any], *args: _P.args, **kwds: _P.kwargs) -> dict[str, Any]: ...

class ClosureVars(NamedTuple):
    nonlocals: Mapping[str, Any]
    globals: Mapping[str, Any]
    builtins: Mapping[str, Any]
    unbound: AbstractSet[str]

def getclosurevars(func: Callable[..., Any]) -> ClosureVars: ...
def unwrap(func: Callable[..., Any], *, stop: Callable[[Any], Any] | None = ...) -> Any: ...

#
# The interpreter stack
#

class Traceback(NamedTuple):
    filename: str
    lineno: int
    function: str
    code_context: list[str] | None
    index: int | None  # type: ignore[assignment]

class FrameInfo(NamedTuple):
    frame: FrameType
    filename: str
    lineno: int
    function: str
    code_context: list[str] | None
    index: int | None  # type: ignore[assignment]

def getframeinfo(frame: FrameType | TracebackType, context: int = ...) -> Traceback: ...
def getouterframes(frame: Any, context: int = ...) -> list[FrameInfo]: ...
def getinnerframes(tb: TracebackType, context: int = ...) -> list[FrameInfo]: ...
def getlineno(frame: FrameType) -> int: ...
def currentframe() -> FrameType | None: ...
def stack(context: int = ...) -> list[FrameInfo]: ...
def trace(context: int = ...) -> list[FrameInfo]: ...

#
# Fetching attributes statically
#

def getattr_static(obj: object, attr: str, default: Any | None = ...) -> Any: ...

#
# Current State of Generators and Coroutines
#

GEN_CREATED: Literal["GEN_CREATED"]
GEN_RUNNING: Literal["GEN_RUNNING"]
GEN_SUSPENDED: Literal["GEN_SUSPENDED"]
GEN_CLOSED: Literal["GEN_CLOSED"]

def getgeneratorstate(
    generator: Generator[Any, Any, Any]
) -> Literal["GEN_CREATED", "GEN_RUNNING", "GEN_SUSPENDED", "GEN_CLOSED"]: ...

CORO_CREATED: Literal["CORO_CREATED"]
CORO_RUNNING: Literal["CORO_RUNNING"]
CORO_SUSPENDED: Literal["CORO_SUSPENDED"]
CORO_CLOSED: Literal["CORO_CLOSED"]

def getcoroutinestate(
    coroutine: Coroutine[Any, Any, Any]
) -> Literal["CORO_CREATED", "CORO_RUNNING", "CORO_SUSPENDED", "CORO_CLOSED"]: ...
def getgeneratorlocals(generator: Generator[Any, Any, Any]) -> dict[str, Any]: ...
def getcoroutinelocals(coroutine: Coroutine[Any, Any, Any]) -> dict[str, Any]: ...

# Create private type alias to avoid conflict with symbol of same
# name created in Attribute class.
_Object = object

class Attribute(NamedTuple):
    name: str
    kind: str
    defining_class: type
    object: _Object

def classify_class_attrs(cls: type) -> list[Attribute]: ...

if sys.version_info >= (3, 9):
    class ClassFoundException(Exception): ...



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Dependency used for managing localising strings around interactions commands."""
from __future__ import annotations

__all__: list[str] = ["AbstractLocaliser", "AbstractLocalizer", "BasicLocaliser", "BasicLocalizer"]

import abc
import typing

if typing.TYPE_CHECKING:
    from collections import abc as collections

    from typing_extensions import Self

    from .. import abc as tanjun


class AbstractLocaliser(abc.ABC):
    """Abstract class of a string localiser."""

    __slots__ = ()

    @abc.abstractmethod
    def get_all_variants(self, identifier: str, /, **kwargs: typing.Any) -> collections.Mapping[str, str]:
        """Get all the localisation variants for an identifier."""

    @abc.abstractmethod
    def localise(self, identifier: str, tag: str, /, **kwargs: typing.Any) -> typing.Optional[str]:
        """Localise a string with the given identifier and arguments.

        Parameters
        ----------
        identifier
            The unique identifier of the string to localise.

            This may be in any format but the formats used by the standard
            implementations can be found at [client-localiser][].
        tag
            The "IETF lang tag" to localise the string to.

            This should usually be a [hikari.locales.Locale][].
        **kwargs
            Key-word arguments to pass to the string as format args.

        Returns
        -------
        str
            The localised string.
        """

    def localize(self, identifier: str, tag: str, /, **kwargs: typing.Any) -> typing.Optional[str]:
        """Alias for `AbstractLocaliser.localise`."""
        return self.localise(identifier, tag, **kwargs)


AbstractLocalizer = AbstractLocaliser
"""Alias of `AbstractLocaliser`."""


class BasicLocaliser(AbstractLocaliser):
    """Standard implementation of `AbstractLocaliser` with only basic text mapping support."""

    __slots__ = ("_tags",)

    def __init__(self) -> None:
        """Initialise a new `BasicLocaliser`."""
        self._tags: dict[str, dict[str, str]] = {}

    def add_to_client(self, client: tanjun.Client, /) -> None:
        """Add this global localiser to a tanjun client.

        !!! note
            This registers the manager as a type dependency to let Tanjun use it.

        Parameters
        ----------
        client
            The client to add this global localiser to.
        """
        client.set_type_dependency(AbstractLocalizer, self)

    def get_all_variants(self, identifier: str, /, **kwargs: typing.Any) -> collections.Mapping[str, str]:
        # <<inherited docstring from AbstractLocaliser>>.
        try:
            results = self._tags[identifier]

        except KeyError:
            return {}

        else:
            if kwargs:
                results = {name: value.format(**kwargs) for name, value in results.items()}

            else:
                results = results.copy()

            results.pop("default", None)
            return results

    def localise(self, identifier: str, tag: str, /, **kwargs: typing.Any) -> typing.Optional[str]:
        # <<inherited docstring from AbstractLocaliser>>.
        if (tag_values := self._tags.get(identifier)) and (string := tag_values.get(tag)):
            return string.format(**kwargs)

        return None  # MyPy compat

    def set_variants(
        self, identifier: str, variants: typing.Optional[collections.Mapping[str, str]] = None, /, **other_variants: str
    ) -> Self:
        """Set the variants for a localised field.

        Parameters
        ----------
        identifier
            Identifier of the field to set the localised variants for.

            This may be in any format but the formats used by the standard
            implementations can be found at [client-localiser][].
        variants
            Mapping of [hikari.locales.Locale][]s to the localised values.

        Returns
        -------
        Self
            The localiser object to enable chained calls.
        """
        if variants:
            other_variants.update(variants)

        self._tags[identifier] = other_variants
        return self


BasicLocalizer = BasicLocaliser
"""Alias of `BasicLocaliser`."""



# -*- coding: utf-8 -*-
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""Placeholder for `proto`'s standard implementations including logic for injecting them."""
import typing

import examples.config
import tanjun
from examples import protos


async def connect_to_database(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
    raise NotImplementedError  # this is a stand in for the real implementation which would be imported


class DatabaseImpl:
    def __init__(self) -> None:
        self._conn: typing.Optional[typing.Any] = None

    async def connect(self, config: examples.config.ExampleConfig = tanjun.inject(type=examples.config.ExampleConfig)):
        self._conn = await connect_to_database(password=config.database_password, url=config.database_url)

    async def get_guild_info(self, guild_id: int) -> typing.Optional[protos.GuildConfig]:
        raise NotImplementedError

    async def get_user_info(self, user_id: int) -> typing.Optional[protos.UserInfo]:
        raise NotImplementedError

    async def remove_user(self, user_id: int) -> None:
        raise NotImplementedError



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

import types
import typing
from collections import abc as collections
from unittest import mock

import alluka
import hikari
import pytest
from hikari import traits

import tanjun
from tanjun.context import base as base_context

_T = typing.TypeVar("_T")


def stub_class(
    cls: type[_T],
    /,
    args: collections.Sequence[typing.Any] = (),
    kwargs: typing.Optional[collections.Mapping[str, typing.Any]] = None,
    **namespace: typing.Any,
) -> _T:
    namespace["__slots__"] = ()

    for name in getattr(cls, "__abstractmethods__", None) or ():
        if name not in namespace:
            namespace[name] = mock.MagicMock()

    name = origin.__name__ if (origin := getattr(cls, "__origin__", None)) else cls.__name__
    new_cls = types.new_class(name, (cls,), exec_body=lambda body: body.update(namespace))
    return typing.cast("type[_T]", new_cls)(*args, **kwargs or {})


@pytest.fixture()
def mock_client() -> tanjun.abc.Client:
    return mock.MagicMock(tanjun.abc.Client, rest=mock.AsyncMock(hikari.api.RESTClient))


@pytest.fixture()
def mock_component() -> tanjun.abc.Component:
    return mock.MagicMock(tanjun.abc.Component)


class TestBaseContext:
    @pytest.fixture()
    def context(self, mock_client: mock.Mock) -> base_context.BaseContext:
        return stub_class(base_context.BaseContext, args=(mock_client,))

    def test_cache_property(self, context: tanjun.abc.Context, mock_client: mock.Mock):
        assert context.cache is mock_client.cache

    def test_client_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.client is mock_client

    def test_component_property(self, context: base_context.BaseContext, mock_component: tanjun.abc.Component):
        assert context.component is None

    def test_events_proprety(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.events is mock_client.events

    def test_rest_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.rest is mock_client.rest

    def test_server_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.server is mock_client.server

    def test_shard_property(self, mock_client: mock.Mock):
        mock_shard = mock.Mock()
        mock_client.shards = mock.MagicMock(spec=traits.ShardAware, shard_count=5, shards={2: mock_shard})
        context = stub_class(base_context.BaseContext, guild_id=hikari.Snowflake(123321123312), args=(mock_client,))

        assert context.shard is mock_shard

    def test_shard_property_when_dm(self, mock_client: mock.Mock):
        mock_shard = mock.Mock()
        mock_client.shards = mock.Mock(shards={0: mock_shard})
        context = stub_class(base_context.BaseContext, guild_id=None, args=(mock_client,))

        assert context.shard is mock_shard

    def test_shard_property_when_no_shards(self, context: tanjun.context.MessageContext):
        context._client = mock.Mock(shards=None)

        assert context.shard is None

    def test_shards_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.shards is mock_client.shards

    def test_voice_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.voice is mock_client.voice

    def test_finalise(self, context: base_context.BaseContext):
        context.finalise()
        assert context._final is True

    def test_set_component(self, context: base_context.BaseContext):
        component = mock.Mock()

        assert context.set_component(component) is context

        assert context.component is component
        assert context.get_type_dependency(tanjun.abc.Component) is component

    def test_set_component_when_none_and_previously_set(self, context: base_context.BaseContext):
        assert isinstance(context.injection_client.get_type_dependency, mock.Mock)
        context.injection_client.get_type_dependency.return_value = alluka.abc.UNDEFINED
        mock_component = mock.Mock()
        context.set_component(mock_component)
        context.set_component(None)

        assert context.component is None
        assert context.get_type_dependency(tanjun.abc.Component) is alluka.abc.UNDEFINED

    def test_set_component_when_none(self, context: base_context.BaseContext):
        assert isinstance(context.injection_client.get_type_dependency, mock.Mock)
        context.injection_client.get_type_dependency.return_value = alluka.abc.UNDEFINED
        context.set_component(None)
        context.set_component(None)

        assert context.component is None
        assert context.get_type_dependency(tanjun.abc.Component) is alluka.abc.UNDEFINED

    def test_set_component_when_final(self, context: base_context.BaseContext):
        component = mock.Mock()
        context.finalise()

        with pytest.raises(TypeError):
            context.set_component(component)

        assert context.component is not component

    def test_get_channel(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        mock_client.cache.get_guild_channel.return_value = mock.Mock(hikari.TextableGuildChannel)

        assert context.get_channel() is mock_client.cache.get_guild_channel.return_value

        mock_client.cache.get_guild_channel.assert_called_once_with(context.channel_id)

    def test_get_channel_when_cache_returns_none(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        mock_client.cache.get_guild_channel.return_value = None

        assert context.get_channel() is None

        mock_client.cache.get_guild_channel.assert_called_once_with(context.channel_id)

    def test_get_channel_when_cacheless(self):
        context = stub_class(base_context.BaseContext, guild_id=None, args=(mock.Mock(cache=None),))

        assert context.get_channel() is None

    def test_get_guild(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        assert context.get_guild() is mock_client.cache.get_guild.return_value
        mock_client.cache.get_guild.assert_called_once_with(context.guild_id)

    def test_get_guild_when_cacheless(self):
        context = stub_class(base_context.BaseContext, guild_id=None, args=(mock.Mock(cache=None),))

        assert context.get_guild() is None

    def test_get_guild_when_dm_bound(self):
        mock_client = mock.MagicMock()
        context = stub_class(base_context.BaseContext, guild_id=None, args=(mock_client,))

        assert context.get_guild() is None
        mock_client.cache.get_guild.assert_not_called()

    @pytest.mark.asyncio()
    async def test_fetch_channel(self, context: base_context.BaseContext, mock_client: mock.Mock):
        mock_client.rest.fetch_channel.return_value = mock.Mock(hikari.TextableChannel)

        result = await context.fetch_channel()

        assert result is mock_client.rest.fetch_channel.return_value
        mock_client.rest.fetch_channel.assert_called_once_with(context.channel_id)

    @pytest.mark.asyncio()
    async def test_fetch_guild(self, context: base_context.BaseContext, mock_client: mock.Mock):
        result = await context.fetch_guild()

        assert result is mock_client.rest.fetch_guild.return_value
        mock_client.rest.fetch_guild.assert_called_once_with(context.guild_id)

    @pytest.mark.asyncio()
    async def test_fetch_guild_when_dm_bound(self, mock_client: mock.Mock):
        context = stub_class(base_context.BaseContext, guild_id=None, args=(mock_client,))

        result = await context.fetch_guild()

        assert result is None
        mock_client.rest.fetch_guild.assert_not_called()



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportIncompatibleMethodOverride=none
# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

import typing
from unittest import mock

import hikari
import pytest

import tanjun
from tanjun import _internal


def test_as_message_menu():
    mock_callback = mock.Mock()

    command = tanjun.as_message_menu(
        "eat",
        always_defer=True,
        default_member_permissions=hikari.Permissions(43123),
        default_to_ephemeral=False,
        dm_enabled=False,
        is_global=False,
    )(mock_callback)

    assert isinstance(command, tanjun.MenuCommand)
    assert command.type is hikari.CommandType.MESSAGE
    assert command.name == "eat"
    assert command._always_defer is True
    assert command.default_member_permissions == hikari.Permissions(43123)
    assert command.defaults_to_ephemeral is False
    assert command.is_dm_enabled is False
    assert command.is_global is False
    assert command.callback is mock_callback
    assert command.wrapped_command is None


def test_as_message_menu_with_defaults():
    mock_callback = mock.Mock()

    command = tanjun.as_message_menu("yeet")(mock_callback)

    assert isinstance(command, tanjun.MenuCommand)
    assert command.callback is mock_callback
    assert command.type is hikari.CommandType.MESSAGE
    assert command.name == "yeet"
    assert command._always_defer is False
    assert command.default_member_permissions is None
    assert command.defaults_to_ephemeral is None
    assert command.is_dm_enabled is None
    assert command.is_global is True
    assert command.callback is mock_callback
    assert command.wrapped_command is None


@pytest.mark.parametrize(
    "other_command",
    [
        tanjun.SlashCommand[typing.Any](mock.Mock(), "e", "a"),
        tanjun.MessageCommand[typing.Any](mock.Mock(), "b"),
        tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "a"),
    ],
)
def test_as_message_menu_when_wrapping_command(
    other_command: typing.Union[
        tanjun.SlashCommand[typing.Any], tanjun.MessageCommand[typing.Any], tanjun.MenuCommand[typing.Any, typing.Any]
    ]
):
    command = tanjun.as_message_menu(
        "c",
        always_defer=False,
        default_member_permissions=hikari.Permissions(543123),
        default_to_ephemeral=True,
        dm_enabled=False,
        is_global=True,
    )(other_command)

    assert command._always_defer is False
    assert command.default_member_permissions == 543123
    assert command.defaults_to_ephemeral is True
    assert command.is_dm_enabled is False
    assert command.is_global is True
    assert command.type is hikari.CommandType.MESSAGE
    assert command.callback is other_command.callback
    assert command.wrapped_command is other_command
    assert isinstance(command, tanjun.MenuCommand)


def test_as_user_menu():
    mock_callback = mock.Mock()

    command = tanjun.as_user_menu(
        "uoy",
        always_defer=True,
        default_member_permissions=hikari.Permissions(49494),
        default_to_ephemeral=False,
        dm_enabled=False,
        is_global=False,
    )(mock_callback)

    assert isinstance(command, tanjun.MenuCommand)
    assert command.callback is mock_callback
    assert command.type is hikari.CommandType.USER
    assert command.name == "uoy"
    assert command._always_defer is True
    assert command.default_member_permissions == hikari.Permissions(49494)
    assert command.defaults_to_ephemeral is False
    assert command.is_dm_enabled is False
    assert command.is_global is False
    assert command.callback is mock_callback
    assert command.wrapped_command is None


def test_as_user_menu_with_defaults():
    mock_callback = mock.Mock()

    command = tanjun.as_user_menu("you")(mock_callback)

    assert isinstance(command, tanjun.MenuCommand)
    assert command.type is hikari.CommandType.USER
    assert command.name == "you"
    assert command._always_defer is False
    assert command.default_member_permissions is None
    assert command.defaults_to_ephemeral is None
    assert command.is_dm_enabled is None
    assert command.is_global is True
    assert command.callback is mock_callback
    assert command.wrapped_command is None


@pytest.mark.parametrize(
    "other_command",
    [
        tanjun.SlashCommand[typing.Any](mock.Mock(), "e", "a"),
        tanjun.MessageCommand[typing.Any](mock.Mock(), "b"),
        tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "a"),
    ],
)
def test_as_user_menu_when_wrapping_command(
    other_command: typing.Union[
        tanjun.SlashCommand[typing.Any], tanjun.MessageCommand[typing.Any], tanjun.MenuCommand[typing.Any, typing.Any]
    ]
):
    command = tanjun.as_user_menu(
        "c",
        always_defer=True,
        default_member_permissions=hikari.Permissions(4212312),
        default_to_ephemeral=False,
        dm_enabled=True,
        is_global=False,
    )(other_command)

    assert command._always_defer is True
    assert command.default_member_permissions == 4212312
    assert command.defaults_to_ephemeral is False
    assert command.is_dm_enabled is True
    assert command.is_global is False
    assert command.type is hikari.CommandType.USER
    assert command.callback is other_command.callback
    assert command.wrapped_command is other_command
    assert isinstance(command, tanjun.MenuCommand)


class TestMenuCommand:
    def test__init__when_no_names_provided(self):
        with pytest.raises(RuntimeError, match="No default name given"):
            tanjun.commands.MenuCommand(mock.AsyncMock(), hikari.CommandType.USER, {"id": "idea"})

    def test__init__when_name_too_long(self):
        with pytest.raises(ValueError, match="Name must be less than or equal to 32 characters in length"):
            tanjun.commands.MenuCommand(mock.Mock(), hikari.CommandType.MESSAGE, "x" * 33)

    def test__init__when_localised_name_too_long(self):
        with pytest.raises(ValueError, match="Name must be less than or equal to 32 characters in length"):
            tanjun.commands.MenuCommand(
                mock.Mock(),
                hikari.CommandType.MESSAGE,
                {hikari.Locale.BG: "year", hikari.Locale.DA: "y" * 33, hikari.Locale.DE: "ein", "default": "meow"},
            )

    def test__init__when_localised_default_name_too_long(self):
        with pytest.raises(ValueError, match="Name must be less than or equal to 32 characters in length"):
            tanjun.commands.MenuCommand(
                mock.Mock(),
                hikari.CommandType.MESSAGE,
                {hikari.Locale.BG: "year", hikari.Locale.DE: "ein", "default": "meow" * 9},
            )

    def test__init__when_name_too_short(self):
        with pytest.raises(ValueError, match="Name must be greater than or equal to 1 characters in length"):
            tanjun.commands.MenuCommand(mock.Mock(), hikari.CommandType.MESSAGE, "")

    def test__init__when_localised_name_too_short(self):
        with pytest.raises(ValueError, match="Name must be greater than or equal to 1 characters in length"):
            tanjun.commands.MenuCommand(
                mock.Mock(),
                hikari.CommandType.MESSAGE,
                {hikari.Locale.BG: "", hikari.Locale.DA: "damn", hikari.Locale.HR: "beep"},
            )

    def test__init__when_localised_default_name_too_short(self):
        with pytest.raises(ValueError, match="Name must be greater than or equal to 1 characters in length"):
            tanjun.commands.MenuCommand(
                mock.Mock(),
                hikari.CommandType.MESSAGE,
                {hikari.Locale.EN_GB: "great scott", "default": "", hikari.Locale.EL: "beep", hikari.Locale.DA: "boop"},
            )

    @pytest.mark.parametrize(
        "inner_command",
        [
            tanjun.SlashCommand[typing.Any](mock.Mock(), "a", "b"),
            tanjun.MessageCommand[typing.Any](mock.Mock(), "a"),
            tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "e"),
        ],
    )
    def test___init___when_command_object(
        self,
        inner_command: typing.Union[
            tanjun.SlashCommand[tanjun.abc.CommandCallbackSig],
            tanjun.MessageCommand[tanjun.abc.CommandCallbackSig],
            tanjun.MenuCommand[typing.Any, typing.Any],
        ],
    ):
        assert tanjun.MenuCommand(inner_command, hikari.CommandType.MESSAGE, "woow").callback is inner_command.callback

    @pytest.mark.asyncio()
    async def test_call_dunder_method(self):
        mock_callback: typing.Any = mock.AsyncMock()
        command = tanjun.MenuCommand(mock_callback, hikari.CommandType.MESSAGE, "a")

        await command(123, 321, "ea", b=32)

        mock_callback.assert_awaited_once_with(123, 321, "ea", b=32)

    def test_callback_property(self):
        mock_callback = mock.Mock()
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock_callback, hikari.CommandType.MESSAGE, "a")

        assert command.callback is mock_callback

    def test_default_member_permissions_property(self):
        mock_callback = mock.Mock()
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock_callback, hikari.CommandType.MESSAGE, "a", default_member_permissions=hikari.Permissions(6541231)
        )

        assert command.default_member_permissions == 6541231

    def test_defaults_to_ephemeral_property(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.MESSAGE, "a", default_to_ephemeral=True
        )

        assert command.defaults_to_ephemeral is True

    def test_defaults_to_ephemeral_property_when_unset(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "a")

        assert command.defaults_to_ephemeral is None

    def test_is_dm_enabled_property(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.MESSAGE, "a", dm_enabled=False
        )

        assert command.is_dm_enabled is False

    def test_is_global_property(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.MESSAGE, "a", is_global=False
        )

        assert command.is_global is False

    def test_is_global_property_when_default(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "a")

        assert command.is_global is True

    def test_name_properties(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "uwu")

        assert command.name == "uwu"
        assert command.name_localisations == {}
        assert command._names.id is None

    def test_name_properties_when_localised(self):
        command = tanjun.MenuCommand(
            mock.AsyncMock(),
            hikari.CommandType.USER,
            {
                hikari.Locale.BG: "hi",
                hikari.Locale.CS: "Nay",
                "default": "nay",
                "id": "meow",
                hikari.Locale.PT_BR: "yeet",
            },
        )

        assert command.name == "nay"
        assert command.name_localisations == {
            hikari.Locale.BG: "hi",
            hikari.Locale.CS: "Nay",
            hikari.Locale.PT_BR: "yeet",
        }
        assert command._names.id == "meow"

    def test_name_properties_when_localised_implicit_default(self):
        command = tanjun.MenuCommand(
            mock.AsyncMock(), hikari.CommandType.MESSAGE, {hikari.Locale.JA: "Rei", hikari.Locale.DE: "Meow"}
        )

        assert command.name == "Rei"
        assert command.name_localisations == {hikari.Locale.JA: "Rei", hikari.Locale.DE: "Meow"}
        assert command._names.id is None

    def test_name_properties_when_dict_without_localisations(self):
        command = tanjun.MenuCommand(
            mock.AsyncMock(), hikari.CommandType.MESSAGE, {"id": "shinjis_man", "default": "no papa"}
        )

        assert command.name == "no papa"
        assert command.name_localisations == {}
        assert command._names.id == "shinjis_man"

    def test_tracked_command_property(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "uwu")

        assert command.tracked_command is None

    def test_tracked_command_id_property(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "uwu")

        assert command.tracked_command_id is None

    @pytest.mark.parametrize("command_type", [hikari.CommandType.MESSAGE, hikari.CommandType.USER])
    def test_type_property(self, command_type: hikari.CommandType):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), command_type, "uwu")  # type: ignore

        assert command.type is command_type

    def test_build(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.USER, "owo")

        builder = command.build()

        assert builder.name == "owo"
        assert builder.name_localizations == {}
        assert builder.type is hikari.CommandType.USER
        assert builder.id is hikari.UNDEFINED
        assert builder.default_member_permissions is hikari.UNDEFINED
        assert builder.is_dm_enabled is hikari.UNDEFINED

    def test_build_when_all_fields_set(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(),
            hikari.CommandType.MESSAGE,
            "pat",
            default_member_permissions=hikari.Permissions(4123),
            dm_enabled=False,
        ).bind_component(
            mock.Mock(default_app_cmd_permissions=hikari.Permissions(54123123), dms_enabled_for_app_cmds=True)
        )

        builder = command.build(
            component=mock.Mock(default_app_cmd_permissions=hikari.Permissions(341123), dms_enabled_for_app_cmds=True)
        )

        assert builder.name == "pat"
        assert builder.type is hikari.CommandType.MESSAGE
        assert builder.id is hikari.UNDEFINED
        assert builder.default_member_permissions == hikari.Permissions(4123)
        assert builder.is_dm_enabled is False
        assert builder.name_localizations == {}

    def test_build_with_localised_fields(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(),
            hikari.CommandType.USER,
            {hikari.Locale.EN_GB: "yeet", hikari.Locale.FI: "beat", "default": "shinji", hikari.Locale.JA: "Ayanami"},
        )

        builder = command.build()

        assert builder.name == "shinji"
        assert builder.name_localizations == {
            hikari.Locale.EN_GB: "yeet",
            hikari.Locale.FI: "beat",
            hikari.Locale.JA: "Ayanami",
        }

    def test_build_with_localised_fields_and_implicit_default(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(),
            hikari.CommandType.USER,
            {
                hikari.Locale.EN_GB: "yeet",
                hikari.Locale.FI: "beat",
                hikari.Locale.DA: "drum",
                hikari.Locale.JA: "Ayanami",
            },
        )

        builder = command.build()

        assert builder.name == "yeet"
        assert builder.name_localizations == {
            hikari.Locale.EN_GB: "yeet",
            hikari.Locale.FI: "beat",
            hikari.Locale.DA: "drum",
            hikari.Locale.JA: "Ayanami",
        }

    def test_build_with_bound_component_field_inheritance(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.USER, "owo"
        ).bind_component(
            mock.Mock(default_app_cmd_permissions=hikari.Permissions(65412312), dms_enabled_for_app_cmds=True)
        )

        builder = command.build()

        assert builder.name == "owo"
        assert builder.type is hikari.CommandType.USER
        assert builder.id is hikari.UNDEFINED
        assert builder.default_member_permissions == 65412312
        assert builder.is_dm_enabled is True

    def test_build_with_passed_component_field_inheritance(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.USER, "owo"
        ).bind_component(
            mock.Mock(default_app_cmd_permissions=hikari.Permissions(65412312), dms_enabled_for_app_cmds=True)
        )

        builder = command.build(
            component=mock.Mock(
                default_app_cmd_permissions=hikari.Permissions(561234123), dms_enabled_for_app_cmds=False
            )
        )

        assert builder.name == "owo"
        assert builder.type is hikari.CommandType.USER
        assert builder.id is hikari.UNDEFINED
        assert builder.default_member_permissions == 561234123
        assert builder.is_dm_enabled is False

    def test_set_tracked_command(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "pat")
        mock_command = mock.Mock(hikari.ContextMenuCommand)

        result = command.set_tracked_command(mock_command)

        assert result is command
        assert command.tracked_command is mock_command
        assert command.tracked_command_id is mock_command.id

    def test_set_ephemeral_default(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "pat")

        result = command.set_ephemeral_default(True)

        assert result is command
        assert command.defaults_to_ephemeral is True

    @pytest.mark.asyncio()
    async def test_check_context(self):
        mock_callback = mock.Mock()
        mock_other_callback = mock.Mock()
        mock_context = mock.Mock()

        command = (
            tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.USER, "pat")
            .add_check(mock_callback)
            .add_check(mock_other_callback)
        )

        with mock.patch.object(_internal, "gather_checks", new=mock.AsyncMock()) as gather_checks:
            result = await command.check_context(mock_context)

            gather_checks.assert_awaited_once_with(mock_context, [mock_callback, mock_other_callback])

        assert result is gather_checks.return_value
        mock_context.set_command.assert_has_calls([mock.call(command), mock.call(None)])

    @pytest.mark.skip(reason="TODO")
    def test_copy(self):
        ...

    @pytest.mark.skip(reason="TODO")
    @pytest.mark.asyncio()
    async def test_execute(self):
        ...

    def test_load_into_component(self):
        command = tanjun.MenuCommand[typing.Any, typing.Any](mock.Mock(), hikari.CommandType.MESSAGE, "pat")
        mock_component = mock.Mock()

        command.load_into_component(mock_component)

        mock_component.add_menu_command.assert_called_once_with(command)

    def test_load_into_component_when_wrapped_command(self):
        mock_other_command = mock.Mock()
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.MESSAGE, "pat", _wrapped_command=mock_other_command
        )
        mock_component = mock.Mock()

        command.load_into_component(mock_component)

        mock_component.add_menu_command.assert_called_once_with(command)
        mock_other_command.load_into_component.assert_not_called()

    def test_load_into_component_when_wrapped_command_is_loader(self):
        mock_other_command = mock.Mock(tanjun.components.AbstractComponentLoader)
        command = tanjun.MenuCommand[typing.Any, typing.Any](
            mock.Mock(), hikari.CommandType.MESSAGE, "pat", _wrapped_command=mock_other_command
        )
        mock_component = mock.Mock()

        command.load_into_component(mock_component)

        mock_component.add_menu_command.assert_called_once_with(command)
        mock_other_command.load_into_component.assert_called_once_with(mock_component)



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Default dependency utilities used within Tanjun and their abstract interfaces."""
from __future__ import annotations

__all__: list[str] = [
    "AbstractConcurrencyLimiter",
    "AbstractCooldownManager",
    "AbstractLocaliser",
    "AbstractLocalizer",
    "AbstractOwners",
    "AsyncCache",
    "BasicLocaliser",
    "BasicLocalizer",
    "BucketResource",
    "CacheIterator",
    "CacheMissError",
    "ChannelBoundCache",
    "ConcurrencyPostExecution",
    "ConcurrencyPreExecution",
    "CooldownPreExecution",
    "EntryNotFound",
    "GuildBoundCache",
    "HotReloader",
    "InMemoryConcurrencyLimiter",
    "InMemoryCooldownManager",
    "LazyConstant",
    "Owners",
    "SfCache",
    "SfChannelBound",
    "SfGuildBound",
    "SingleStoreCache",
    "async_cache",
    "cached_inject",
    "callbacks",
    "data",
    "fetch_my_user",
    "inject_lc",
    "limiters",
    "locales",
    "owners",
    "reloaders",
    "set_standard_dependencies",
    "with_concurrency_limit",
    "with_cooldown",
]

import typing

import hikari

from .async_cache import AsyncCache
from .async_cache import CacheIterator
from .async_cache import CacheMissError
from .async_cache import ChannelBoundCache
from .async_cache import EntryNotFound
from .async_cache import GuildBoundCache
from .async_cache import SfCache
from .async_cache import SfChannelBound
from .async_cache import SfGuildBound
from .async_cache import SingleStoreCache
from .callbacks import fetch_my_user
from .data import LazyConstant
from .data import cached_inject
from .data import inject_lc
from .limiters import AbstractConcurrencyLimiter
from .limiters import AbstractCooldownManager
from .limiters import BucketResource
from .limiters import ConcurrencyPostExecution
from .limiters import ConcurrencyPreExecution
from .limiters import CooldownPreExecution
from .limiters import InMemoryConcurrencyLimiter
from .limiters import InMemoryCooldownManager
from .limiters import with_concurrency_limit
from .limiters import with_cooldown
from .locales import AbstractLocaliser
from .locales import AbstractLocalizer
from .locales import BasicLocaliser
from .locales import BasicLocalizer
from .owners import AbstractOwners
from .owners import Owners
from .reloaders import HotReloader

if typing.TYPE_CHECKING:
    from .. import abc as _tanjun


def set_standard_dependencies(client: _tanjun.Client, /) -> None:
    """Set the standard dependencies for Tanjun.

    Parameters
    ----------
    client
        The injector client to set the standard dependencies on.
    """
    client.set_type_dependency(AbstractOwners, Owners()).set_type_dependency(
        LazyConstant[hikari.OwnUser], LazyConstant[hikari.OwnUser](fetch_my_user)
    )



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.List[str] = [
    "build",
    "cleanup",
    "copy_actions",
    "flake8",
    "freeze_locks",
    "generate_docs",
    "publish",
    "reformat",
    "slot_check",
    "spell_check",
    "test",
    "test_coverage",
    "test_publish",
    "type_check",
    "verify_deps",
    "verify_markup",
    "verify_types",
]

import datetime
import itertools
import pathlib
import re
import typing
from collections import abc as collections

import nox
import piped_shared
import tomli

_CallbackT = typing.TypeVar("_CallbackT", bound=collections.Callable[..., typing.Any])

_config = piped_shared.Config.read(pathlib.Path("./"))
nox.options.sessions = _config.default_sessions
_DEPS_DIR = pathlib.Path("./dev-requirements")
_SELF_INSTALL_REGEX = re.compile(r"^\.\[.+\]$")


def _tracked_files(session: nox.Session, *, force_all: bool = False) -> collections.Iterable[str]:
    output = session.run("git", "--no-pager", "grep", "--threads=1", "-l", "", external=True, log=False, silent=True)
    assert isinstance(output, str)

    if _config.path_ignore and not force_all:
        return (path for path in output.splitlines() if not _config.path_ignore.search(path))

    return output.splitlines()


def _dev_path(value: str, /) -> typing.Optional[pathlib.Path]:
    for path in [_DEPS_DIR / f"{value}.txt", pathlib.Path(__file__).parent / "base-requirements" / f"{value}.txt"]:
        if path.exists():
            return path

    return None


def _other_dep(name: str) -> str:
    return f"other-{name}"


def _install_deps(
    session: nox.Session, *requirements_tuple: str, names: list[str] = [], first_call: bool = True
) -> None:
    if not requirements_tuple and not names:
        return

    requirements = list(requirements_tuple)
    other_requirements: dict[pathlib.Path, list[str]] = {}
    for name in names:
        if name != "constraints":
            path = _dev_path(name)
            assert path, "If this doesn't exist then path is missing the base requirement and is broken"
            requirements.extend(["-r", str(path)])
            if path := _dev_path(_other_dep(name)):
                files = other_requirements.get(path.parent) or []
                files.extend(["-r", path.name])
                other_requirements[path.parent] = files

            continue

        path = _DEPS_DIR / "constraints.txt"
        if path.exists():
            requirements.extend(["-r", str(path)])

        other_path = path.with_name(_other_dep(path.name))
        if other_path.exists():
            files = other_requirements.get(other_path.parent) or []
            files.extend(["-r", other_path.name])
            other_requirements[other_path.parent] = files

    # --no-install --no-venv leads to it trying to install in the global venv
    # as --no-install only skips "reused" venvs and global is not considered reused.
    if not _try_find_option(session, "--skip-install", when_empty="True"):
        if first_call:
            session.install("--upgrade", "wheel")

        session.install("--upgrade", *requirements)

    elif any(map(_SELF_INSTALL_REGEX.fullmatch, requirements)):
        session.install("--upgrade", "--force-reinstall", "--no-dependencies", ".")

    for cwd, paths in other_requirements.items():
        # We override the cwd using --root to handle relative dep links as relative to the requirement file.
        # TODO: do we want to keep no-deps here?
        with session.chdir(cwd):  # TODO: get "--root", str(cwd) to work
            session.install("--upgrade", "--no-deps", *paths)


def _try_find_option(
    session: nox.Session, name: str, /, *other_names: str, when_empty: typing.Optional[str] = None
) -> typing.Optional[str]:
    args_iter = iter(session.posargs)
    names = {name, *other_names}

    for arg in args_iter:
        if arg in names:
            return next(args_iter, when_empty)


def _filtered_session(
    *,
    python: typing.Union[str, collections.Sequence[str], bool, None] = None,
    py: typing.Union[str, collections.Sequence[str], bool, None] = None,
    reuse_venv: typing.Optional[bool] = None,
    name: typing.Optional[str] = None,
    venv_backend: typing.Any = None,
    venv_params: typing.Any = None,
    tags: typing.Optional[list[str]] = None,
) -> collections.Callable[[_CallbackT], typing.Union[_CallbackT, None]]:
    """Filtering version of `nox.session`."""

    def decorator(callback: _CallbackT, /) -> typing.Optional[_CallbackT]:
        name_ = name or callback.__name__
        if name_ in _config.hide:
            return None

        return nox.session(
            python=python,
            py=py,
            reuse_venv=reuse_venv,
            name=name,
            venv_backend=venv_backend,
            venv_params=venv_params,
            tags=tags,
        )(callback)

    return decorator


@_filtered_session(venv_backend="none")
def cleanup(session: nox.Session) -> None:
    """Cleanup any temporary files made in this project by its nox tasks."""
    import shutil

    # Remove directories
    raw_paths = ["./dist", "./site", "./.nox", "./.pytest_cache", "./coverage_html", ".mypy_cache"]
    if _config.project_name:
        raw_paths.append(f"{_config.project_name}.egg-info")

    for raw_path in raw_paths:
        path = pathlib.Path(raw_path)
        try:
            shutil.rmtree(str(path.absolute()))

        except FileNotFoundError:
            session.warn(f"[ SKIP ] '{raw_path}'")

        except Exception as exc:
            session.error(f"[ FAIL ] Failed to remove '{raw_path}': {exc!s}")  # noqa: TC400

        else:
            session.log(f"[  OK  ] Removed '{raw_path}'")

    # Remove individual files
    for raw_path in ["./.coverage", "./coverage_html.xml", "./gogo.patch"]:
        path = pathlib.Path(raw_path)
        try:
            path.unlink()

        except FileNotFoundError:
            session.warn(f"[ SKIP ] '{raw_path}'")

        except Exception as exc:
            session.error(f"[ FAIL ] Failed to remove '{raw_path}': {exc!s}")  # noqa: TC400

        else:
            session.log(f"[  OK  ] Removed '{raw_path}'")


_DEFAULT_COMMITER_USERNAME = "always-on-duty[bot]"

_ACTION_DEFAULTS = {
    "ACTION_COMMITTER_EMAIL": f"120557446+{_DEFAULT_COMMITER_USERNAME}@users.noreply.github.com",
    "ACTION_COMMITTER_USERNAME": _DEFAULT_COMMITER_USERNAME,
    "DEFAULT_PY_VER": "3.9",
    "NOX_DEP_PATH": "./piped/python/base-requirements/nox.txt",
}
_resync_filter: list[str] = ["piped", "pyproject.toml", "requirements.in"]
_verify_filter: list[str] = []
_dep_locks: list[pathlib.Path] = []


if _config.dep_locks:
    for _path in _config.dep_locks:
        if _path.is_absolute():
            _path = _path.relative_to(pathlib.Path.cwd())

        _str_path = str(_path).replace("\\", "/").strip("/")

        if _path.is_file():
            _resync_filter.append(_str_path)
            _verify_filter.append(_str_path.replace(".in", ".txt"))
            _dep_locks.append(_path.with_name(_path.name.replace(".in", ".txt")))

        else:
            _resync_filter.append(f"{_str_path}/*.in")
            _resync_filter.append(f"!{_str_path}/constraints.in")
            _verify_filter.append(f"{_str_path}/*.txt")
            _dep_locks.append(_path)


class _Action:
    __slots__ = ("defaults", "required_names")

    def __init__(
        self, *, required: collections.Sequence[str] = (), defaults: typing.Optional[piped_shared.ConfigT] = None
    ) -> None:
        self.defaults: piped_shared.ConfigT = dict(_ACTION_DEFAULTS)
        self.defaults.update(defaults or ())
        self.required_names = frozenset(required or ())

    def process_config(self, config: piped_shared.ConfigT, /) -> piped_shared.ConfigT:
        output: piped_shared.ConfigT = dict(self.defaults)
        output.update(**config)
        return output


_ACTIONS: dict[str, _Action] = {
    "clippy": _Action(),
    "freeze-for-pr": _Action(defaults={"EXTEND_FILTERS": [], "FILTERS": _resync_filter}),
    "lint": _Action(),
    "pr-docs": _Action(),
    "publish": _Action(),
    "py-test": _Action(
        required=["PYTHON_VERSIONS"],
        defaults={"CODECLIMATE_TOKEN": "", "OSES": "[ubuntu-latest, macos-latest, windows-latest]"},
    ),
    "reformat": _Action(),
    "release-docs": _Action(defaults={"BRANCH_PUSHES": None}),
    "resync-piped": _Action(defaults={"FILTERS": ["piped", "pyproject.toml"]}),
    "rustfmt": _Action(),
    "type-check": _Action(),
    "update-licence": _Action(),
    "upgrade-locks": _Action(),
    "verify-locks": _Action(defaults={"EXTEND_FILTERS": [], "FILTERS": _verify_filter}),
    "verify-types": _Action(),
}


@nox.session(name="copy-actions")
def copy_actions(_: nox.Session) -> None:
    """Copy over the github actions from Piped without updating the git reference."""
    import jinja2

    env = jinja2.Environment(  # noqa: S701
        keep_trailing_newline=True,
        loader=jinja2.FileSystemLoader(pathlib.Path(__file__).parent.parent / "github" / "actions"),
    )

    env.filters["quoted"] = '"{}"'.format  # noqa: FS002

    to_write: dict[pathlib.Path, str] = {}
    if isinstance(_config.github_actions, dict):
        actions = iter(_config.github_actions.items())
        wild_card: collections.ItemsView[str, piped_shared.ConfigEntryT] = (
            _config.github_actions.get("*") or {}
        ).items()

    else:
        actions: collections.Iterable[tuple[str, piped_shared.ConfigT]] = (
            (name, {}) for name in _config.github_actions
        )
        wild_card = {}.items()

    for file_name, config in actions:
        if file_name == "*":
            continue

        config = {key.upper(): value for key, value in itertools.chain(wild_card, config.items())}
        file_name = file_name.replace("_", "-")
        action = _ACTIONS[file_name]
        if missing := action.required_names.difference(config.keys()):
            raise RuntimeError(f"Missing the following required fields for {file_name} actions: " + ", ".join(missing))

        file_name = f"{file_name}.yml"
        template = env.get_template(file_name)

        full_config = action.process_config(config)
        to_write[pathlib.Path("./.github/workflows") / file_name] = template.render(**full_config, config=_config)

    pathlib.Path("./.github/workflows").mkdir(exist_ok=True, parents=True)

    for path, value in to_write.items():
        with path.open("w+") as file:
            file.write(value)


def _pyproject_toml() -> typing.Optional[dict[str, typing.Any]]:
    try:
        with pathlib.Path("pyproject.toml").open("rb") as _file:
            return tomli.load(_file)

    except FileNotFoundError:
        return None


def _freeze_file(session: nox.Session, path: pathlib.Path, /) -> None:
    if not path.is_symlink():
        target = path.with_name(path.name[:-3] + ".txt")
        target.unlink(missing_ok=True)
        session.run(
            "pip-compile-cross-platform",
            "-o",
            str(target),
            "--min-python-version",
            _config.version(_pyproject_toml()),
            str(path),
        )


_EXTRAS_FILTER = re.compile(r"\[.+\]")


@nox.session(name="freeze-locks", reuse_venv=True)
def freeze_locks(session: nox.Session) -> None:
    """Freeze the dependency locks."""
    _install_deps(session, names=["freeze-locks"])
    constraints_lock = pathlib.Path("./dev-requirements/constraints.txt")
    project = _pyproject_toml()

    if not project:
        raise RuntimeError("Missing pyproject.toml")

    project = project.get("project") or {}
    deps: list[str] = project.get("dependencies") or []
    if optional := project.get("optional-dependencies"):
        deps.extend(itertools.chain(*optional.values()))

    tl_requirements = pathlib.Path("./requirements.in")
    if tl_requirements.exists():
        with tl_requirements.open("r") as file:
            deps.extend(file.read().splitlines())

    constraints_in = pathlib.Path("./dev-requirements/constraints.in")
    if deps:
        with constraints_in.open("w+") as file:
            file.write("\n".join(deps) + "\n")

    else:
        constraints_in.unlink(missing_ok=True)
        constraints_lock.unlink(missing_ok=True)

    for lock_path in _config.dep_locks:
        if lock_path.is_file():
            _freeze_file(session, lock_path)
            continue

        for path in itertools.filterfalse(pathlib.Path.is_symlink, lock_path.glob("*.in")):
            _freeze_file(session, path)

    # We use this file as a constraints file and constraint files can't have extra
    # params according to the spec so these must be removed.
    # See https://github.com/pypa/pip/issues/8210 for more information.
    if constraints_lock.exists():
        with constraints_lock.open("r") as file:
            constraints_value = file.read()

        with constraints_lock.open("w") as file:
            file.write(_EXTRAS_FILTER.sub("", constraints_value))


@_filtered_session(name="verify-deps", reuse_venv=True)
def verify_deps(session: nox.Session) -> None:
    """Verify the dependency locks by installing them."""
    for path in itertools.chain.from_iterable(path.glob("*.txt") if path.is_dir() else (path,) for path in _dep_locks):
        # We override the cwd using --root to handle relative dep links as relative to the requirement file.
        # TODO: do we want to keep no-deps here?
        with session.chdir(path.parent):  # TODO: get "--root", str(cwd) to work
            session.install("--dry-run", "-r", str(path.relative_to(path.parent)))


@_filtered_session(name="generate-docs", reuse_venv=True)
def generate_docs(session: nox.Session) -> None:
    """Generate docs for this project using Mkdoc."""
    _install_deps(session, names=["docs"])
    output_directory = _try_find_option(session, "-o", "--output") or "./site"
    session.run("mkdocs", "build", "-d", output_directory)


@_filtered_session(reuse_venv=True)
def flake8(session: nox.Session) -> None:
    """Run this project's modules against the pre-defined flake8 linters."""
    _install_deps(session, names=["flake8"])
    session.log("Running flake8")
    session.run("pflake8", *_config.top_level_targets, log=False)


@_filtered_session(reuse_venv=True, name="slot-check")
def slot_check(session: nox.Session) -> None:
    """Check this project's slotted classes for common mistakes."""
    # TODO: don't require installing .?
    # https://github.com/pypa/pip/issues/10362
    _install_deps(session, names=["constraints", "lint"])
    _install_deps(session, "--no-deps", *_config.extra_test_installs, first_call=False)
    session.run("slotscheck", "-m", _config.assert_project_name())


@_filtered_session(reuse_venv=True, name="spell-check")
def spell_check(session: nox.Session) -> None:
    """Check this project's text-like files for common spelling mistakes."""
    _install_deps(session, names=["lint"])
    session.log("Running codespell")
    session.run("codespell", *_tracked_files(session), *_config.codespell_ignore_args(), log=False)


@_filtered_session(reuse_venv=True)
def build(session: nox.Session) -> None:
    """Build this project using flit."""
    _install_deps(session, names=["publish"])
    session.log("Starting build")
    session.run("flit", "build")


_LICENCE_PATTERN = re.compile(r"(Copyright \(c\) (\d+-?\d*))")


def _update_licence(match: re.Match[str]) -> str:
    licence_str, date_range = match.groups()
    start = date_range.split("-", 1)[0]
    current_year = str(datetime.datetime.now().year)

    if start == current_year:
        return licence_str

    return f"{licence_str.removesuffix(date_range)}{start}-{current_year}"


_LICENCE_FILE_PATTERN = re.compile(r".*(rs|py)|LICENSE")


@_filtered_session(name="update-licence", venv_backend="none")
def update_licence(session: nox.Session):
    """Bump the end year of the project's licence to the current year."""
    for path in map(pathlib.Path, _tracked_files(session)):
        if not _LICENCE_FILE_PATTERN.fullmatch(path.name):
            continue

        data = path.read_text()
        new_data = _LICENCE_PATTERN.sub(_update_licence, data)

        if new_data != data:
            path.write_text(new_data)


@_filtered_session(name="verify-markup", reuse_venv=True)
def verify_markup(session: nox.Session):
    """Verify the syntax of the repo's markup files."""
    _install_deps(session, names=["lint"])
    tracked_files = list(_tracked_files(session))

    session.log("Running pre_commit_hooks.check_toml")
    session.run(
        "python",
        "-m",
        "pre_commit_hooks.check_toml",
        *(path for path in tracked_files if path.endswith(".toml")),
        success_codes=[0, 1],
        log=False,
    )

    session.log("Running pre_commit_hooks.check_yaml")
    session.run(
        "python",
        "-m",
        "pre_commit_hooks.check_yaml",
        *(path for path in tracked_files if path.endswith(".yml") or path.endswith(".yaml")),
        success_codes=[0, 1],
        log=False,
    )


def _publish(session: nox.Session, /, *, env: typing.Optional[dict[str, str]] = None) -> None:
    # https://github.com/pypa/pip/issues/10362
    _install_deps(session, names=["publish"])

    env = env or session.env.copy()
    if target := session.env.get("PUBLISH_TARGET"):
        env["FLIT_INDEX_URL"] = target

    if token := session.env.get("PUBLISH_TOKEN"):
        env["FLIT_PASSWORD"] = token

    env.setdefault("FLIT_USERNAME", "__token__")
    session.run("flit", "publish", env=env)


@_filtered_session(reuse_venv=True)
def publish(session: nox.Session):
    """Publish this project to pypi."""
    _publish(session)


@_filtered_session(name="test-publish", reuse_venv=True)
def test_publish(session: nox.Session) -> None:
    """Publish this project to test pypi."""
    env = session.env.copy()
    env.setdefault("PYPI_TARGET", "https://test.pypi.org/legacy/")
    _publish(session, env=env)


@_filtered_session(reuse_venv=True)
def reformat(session: nox.Session) -> None:
    """Reformat this project's modules to fit the standard style."""
    _install_deps(session, names=["reformat"])
    if _config.top_level_targets:
        session.run("black", *_config.top_level_targets)
        session.run("isort", *_config.top_level_targets)
        session.run("pycln", *_config.top_level_targets)

    tracked_files = list(_tracked_files(session))  # TODO: sometimes force all or more granular controls?
    py_files = [path for path in tracked_files if re.fullmatch(r".+\.pyi?$", path)]

    if py_files:
        session.log("Running sort-all")
        session.run("sort-all", *py_files, success_codes=[0, 1], log=False)

    session.log("Running pre_commit_hooks.end_of_file_fixer")
    session.run("python", "-m", "pre_commit_hooks.end_of_file_fixer", *tracked_files, success_codes=[0, 1], log=False)

    session.log("Running pre_commit_hooks.trailing_whitespace_fixer")
    session.run(
        "python", "-m", "pre_commit_hooks.trailing_whitespace_fixer", *tracked_files, success_codes=[0, 1], log=False
    )


@_filtered_session(reuse_venv=True)
def test(session: nox.Session) -> None:
    """Run this project's tests using pytest."""
    # https://github.com/pypa/pip/issues/10362
    _install_deps(session, names=["tests"])
    _install_deps(session, *_config.extra_test_installs, first_call=False)
    # TODO: can import-mode be specified in the config.
    session.run("pytest", "-n", "auto", "--import-mode", "importlib")


@_filtered_session(name="test-coverage", reuse_venv=True)
def test_coverage(session: nox.Session) -> None:
    """Run this project's tests while recording test coverage."""
    project_name = _config.assert_project_name()
    # https://github.com/pypa/pip/issues/10362
    _install_deps(session, names=["tests"])
    _install_deps(session, *_config.extra_test_installs, first_call=False)
    # TODO: can import-mode be specified in the config.
    # https://github.com/nedbat/coveragepy/issues/1002
    session.run(
        "pytest",
        "-n",
        "auto",
        f"--cov={project_name}",
        "--cov-report",
        "html:coverage_html",
        "--cov-report",
        "xml:coverage.xml",
    )


def _run_pyright(session: nox.Session, /, *args: str) -> None:
    session.run("python", "-m", "pyright", "--version")
    session.run("python", "-m", "pyright", *args)


@_filtered_session(name="type-check", reuse_venv=True)
def type_check(session: nox.Session) -> None:
    """Statically analyse and veirfy this project using Pyright."""
    _install_deps(session, names=["type-checking"])
    _run_pyright(session)

    if _config.mypy_targets:
        success_codes = [0]
        if _config.mypy_allowed_to_fail:
            success_codes.append(1)

        session.run("python", "-m", "mypy", *_config.mypy_targets, "--show-error-codes", success_codes=success_codes)


@_filtered_session(name="verify-types", reuse_venv=True)
def verify_types(session: nox.Session) -> None:
    """Verify the "type completeness" of types exported by the library using Pyright."""
    project_name = _config.assert_project_name()
    # TODO is installing . necessary here?
    # https://github.com/pypa/pip/issues/10362
    _install_deps(session, names=["type-checking"])
    _install_deps(session, *_config.extra_test_installs, first_call=False)
    _run_pyright(session, "--verifytypes", project_name, "--ignoreexternal")


@nox.session(name="copy-piped", reuse_venv=True)
def sync_piped(session: nox.Session) -> None:
    """Copy over Piped's configuration without fetching."""
    copy_actions(session)


@_filtered_session(name="fetch-piped", reuse_venv=True)
def fetch_piped(session: nox.Session) -> None:
    """Fetch Piped from upstream and resync."""
    session.run("git", "submodule", "update", "--remote", "piped", external=True)
    _install_deps(session, names=["nox"])
    # We call this through nox's CLI like this to ensure that the updated version
    # of these sessions are called.
    session.run("nox", "-s", "copy-piped")


@nox.session(name="bot-package-diff", venv_backend="none")
def bot_package_diff(session: nox.Session) -> None:
    session.run("git", "add", ".", external=True)
    output = session.run("git", "diff", "HEAD", external=True, silent=True)
    assert isinstance(output, str)

    path = pathlib.Path("./gogo.patch")
    if output:
        with path.open("w+") as file:
            file.write(output)

    else:
        path.unlink(missing_ok=True)


@nox.session(name="is-diff-file-empty", venv_backend="none")
def is_diff_file_empty(_: nox.Session):
    if pathlib.Path("./gogo.patch").exists():
        exit("Diff created")



# -*- coding: utf-8 -*-
# BSD 3-Clause License
#
# Copyright (c) 2020-2023, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Message command implementations."""
from __future__ import annotations

__all__: list[str] = ["MessageCommand", "MessageCommandGroup", "as_message_command", "as_message_command_group"]

import copy
import typing
from collections import abc as collections

from .. import _internal
from .. import abc as tanjun
from .. import components
from .. import errors
from .. import hooks as hooks_
from . import base

if typing.TYPE_CHECKING:
    from typing_extensions import Self

    _AnyMessageCommandT = typing.TypeVar("_AnyMessageCommandT", bound=tanjun.MessageCommand[typing.Any])
    _AnyCallbackSigT = typing.TypeVar("_AnyCallbackSigT", bound=collections.Callable[..., typing.Any])
    _AnyCommandT = typing.Union[
        tanjun.MenuCommand[_AnyCallbackSigT, typing.Any],
        tanjun.MessageCommand[_AnyCallbackSigT],
        tanjun.SlashCommand[_AnyCallbackSigT],
    ]
    _CallbackishT = typing.Union[_AnyCommandT["_MessageCallbackSigT"], "_MessageCallbackSigT"]
    _OtherCallbackSigT = typing.TypeVar("_OtherCallbackSigT", bound=tanjun.MessageCallbackSig)


_MessageCallbackSigT = typing.TypeVar("_MessageCallbackSigT", bound=tanjun.MessageCallbackSig)
_EMPTY_DICT: typing.Final[dict[typing.Any, typing.Any]] = {}
_EMPTY_HOOKS: typing.Final[hooks_.Hooks[typing.Any]] = hooks_.Hooks()


# While these overloads may seem redundant/unnecessary, MyPy cannot understand
# this when expressed through callback: _CallbackIshT[_MessageCallbackSigT].
class _AsMsgResultProto(typing.Protocol):
    @typing.overload
    def __call__(self, _: _MessageCallbackSigT, /) -> MessageCommand[_MessageCallbackSigT]:
        ...

    @typing.overload
    def __call__(self, _: _AnyCommandT[_MessageCallbackSigT], /) -> MessageCommand[_MessageCallbackSigT]:
        ...


def as_message_command(name: str, /, *names: str, validate_arg_keys: bool = True) -> _AsMsgResultProto:
    """Build a message command from a decorated callback.

    Parameters
    ----------
    name
        The command name.
    *names
        Other names for the command.
    validate_arg_keys
        Whether to validate that option keys match the command callback's signature.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCallbackSig], MessageCommand]
        The decorator callback used to make a [tanjun.MessageCommand][].

        This can either wrap a raw command callback or another callable command instance
        (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][])
        and will manage loading the other command into a component when using
        [tanjun.Component.load_from_scope][].
    """

    def decorator(callback: _CallbackishT[_MessageCallbackSigT], /) -> MessageCommand[_MessageCallbackSigT]:
        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            wrapped_command = callback
            callback = callback.callback

        else:
            wrapped_command = None

        return MessageCommand(
            callback, name, *names, validate_arg_keys=validate_arg_keys, _wrapped_command=wrapped_command
        )

    return decorator


# While these overloads may seem redundant/unnecessary, MyPy cannot understand
# this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
class _AsGroupResultProto(typing.Protocol):
    @typing.overload
    def __call__(self, _: _MessageCallbackSigT, /) -> MessageCommandGroup[_MessageCallbackSigT]:
        ...

    @typing.overload
    def __call__(self, _: _AnyCommandT[_MessageCallbackSigT], /) -> MessageCommandGroup[_MessageCallbackSigT]:
        ...


def as_message_command_group(
    name: str, /, *names: str, strict: bool = False, validate_arg_keys: bool = True
) -> _AsGroupResultProto:
    """Build a message command group from a decorated callback.

    Parameters
    ----------
    name
        The command name.
    *names
        Other names for the command.
    strict
        Whether this command group should only allow commands without spaces in their names.

        This allows for a more optimised command search pattern to be used and
        enforces that command names are unique to a single command within the group.
    validate_arg_keys
        Whether to validate that option keys match the command callback's signature.

    Returns
    -------
    collections.abc.Callable[[tanjun.abc.MessageCallbackSig], MessageCommand]
        The decorator callback used to make a [tanjun.MessageCommandGroup][].

        This can either wrap a raw command callback or another callable command instance
        (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][])
        and will manage loading the other command into a component when using
        [tanjun.Component.load_from_scope][].
    """

    def decorator(callback: _CallbackishT[_MessageCallbackSigT], /) -> MessageCommandGroup[_MessageCallbackSigT]:
        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            wrapped_command = callback
            callback = callback.callback

        else:
            wrapped_command = None

        return MessageCommandGroup(
            callback, name, *names, strict=strict, validate_arg_keys=validate_arg_keys, _wrapped_command=wrapped_command
        )

    return decorator


class MessageCommand(base.PartialCommand[tanjun.MessageContext], tanjun.MessageCommand[_MessageCallbackSigT]):
    """Standard implementation of a message command."""

    __slots__ = ("_arg_names", "_callback", "_names", "_parent", "_parser", "_wrapped_command")

    # While these overloads may seem redundant/unnecessary, MyPy cannot understand
    # this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
    @typing.overload
    def __init__(
        self,
        callback: _MessageCallbackSigT,
        name: str,
        /,
        *names: str,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        callback: _AnyCommandT[_MessageCallbackSigT],
        name: str,
        /,
        *names: str,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    def __init__(
        self,
        callback: _CallbackishT[_MessageCallbackSigT],
        name: str,
        /,
        *names: str,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        """Initialise a message command.

        Parameters
        ----------
        callback : tanjun.abc.MessageCallbackSig
            Callback to execute when the command is invoked.

            This should be an asynchronous callback which takes one positional
            argument of type [tanjun.abc.MessageContext][], returns [None][]
            and may use dependency injection to access other services.
        name
            The command name.
        *names
            Other names for the command.
        validate_arg_keys
            Whether to validate that option keys match the command callback's signature.
        """
        super().__init__()
        if isinstance(callback, (tanjun.MenuCommand, tanjun.MessageCommand, tanjun.SlashCommand)):
            callback = callback.callback

        self._arg_names = _internal.get_kwargs(callback) if validate_arg_keys else None
        self._callback: _MessageCallbackSigT = callback
        self._names = list(dict.fromkeys((name, *names)))
        self._parent: typing.Optional[tanjun.MessageCommandGroup[typing.Any]] = None
        self._parser: typing.Optional[tanjun.MessageParser] = None
        self._wrapped_command = _wrapped_command

    def __repr__(self) -> str:
        return f"Command <{self._names}>"

    if typing.TYPE_CHECKING:
        __call__: _MessageCallbackSigT  # pyright: ignore [ reportUninitializedInstanceVariable ]

    else:

        async def __call__(self, *args, **kwargs) -> None:
            await self._callback(*args, **kwargs)

    @property
    def callback(self) -> _MessageCallbackSigT:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        return self._callback

    @property
    # <<inherited docstring from tanjun.abc.MessageCommand>>.
    def names(self) -> collections.Collection[str]:
        return self._names.copy()

    @property
    def parent(self) -> typing.Optional[tanjun.MessageCommandGroup[typing.Any]]:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        return self._parent

    @property
    def parser(self) -> typing.Optional[tanjun.MessageParser]:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        return self._parser

    @property
    def wrapped_command(self) -> typing.Optional[tanjun.ExecutableCommand[typing.Any]]:
        """The command object this wraps, if any."""
        return self._wrapped_command

    def bind_client(self, client: tanjun.Client, /) -> Self:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        super().bind_client(client)
        if self._parser:
            self._parser.bind_client(client)

        return self

    def bind_component(self, component: tanjun.Component, /) -> Self:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        super().bind_component(component)
        if self._parser:
            self._parser.bind_component(component)

        return self

    def copy(self, *, parent: typing.Optional[tanjun.MessageCommandGroup[typing.Any]] = None) -> Self:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        inst = super().copy()
        inst._callback = copy.copy(self._callback)
        inst._names = self._names.copy()
        inst._parent = parent
        inst._parser = self._parser.copy() if self._parser else None
        return inst

    def set_parent(self, parent: typing.Optional[tanjun.MessageCommandGroup[typing.Any]], /) -> Self:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        self._parent = parent
        return self

    def set_parser(self, parser: typing.Optional[tanjun.MessageParser], /) -> Self:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        if parser and self._arg_names is not None:
            try:
                name = self.callback.__name__
            except AttributeError:  # Not all callables have names
                name = repr(self.callback)

            parser.validate_arg_keys(name, self._arg_names)

        self._parser = parser
        return self

    async def check_context(self, ctx: tanjun.MessageContext, /) -> bool:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        ctx.set_command(self)
        result = await _internal.gather_checks(ctx, self._checks)
        ctx.set_command(None)
        return result

    async def execute(
        self,
        ctx: tanjun.MessageContext,
        /,
        *,
        hooks: typing.Optional[collections.MutableSet[tanjun.MessageHooks]] = None,
    ) -> None:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        ctx = ctx.set_command(self)
        own_hooks = self._hooks or _EMPTY_HOOKS
        try:
            await own_hooks.trigger_pre_execution(ctx, hooks=hooks)

            if self._parser is not None:
                kwargs = await self._parser.parse(ctx)

            else:
                kwargs = _EMPTY_DICT

            await ctx.call_with_async_di(self._callback, ctx, **kwargs)

        except errors.CommandError as exc:
            await exc.send(ctx)

        except errors.HaltExecution:
            raise

        except Exception as exc:
            if await own_hooks.trigger_error(ctx, exc, hooks=hooks) <= 0:
                raise

        else:
            # TODO: how should this be handled around CommandError?
            await own_hooks.trigger_success(ctx, hooks=hooks)

        finally:
            await own_hooks.trigger_post_execution(ctx, hooks=hooks)

    def load_into_component(self, component: tanjun.Component, /) -> None:
        # <<inherited docstring from tanjun.components.load_into_component>>.
        if not self._parent:
            component.add_message_command(self)

        if self._wrapped_command and isinstance(self._wrapped_command, components.AbstractComponentLoader):
            self._wrapped_command.load_into_component(component)


class MessageCommandGroup(MessageCommand[_MessageCallbackSigT], tanjun.MessageCommandGroup[_MessageCallbackSigT]):
    """Standard implementation of a message command group."""

    __slots__ = ("_commands",)

    # While these overloads may seem redundant/unnecessary, MyPy cannot understand
    # this when expressed through `callback: _CallbackIshT[_MessageCallbackSigT]`.
    @typing.overload
    def __init__(
        self,
        callback: _MessageCallbackSigT,
        name: str,
        /,
        *names: str,
        strict: bool = False,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        callback: _AnyCommandT[_MessageCallbackSigT],
        name: str,
        /,
        *names: str,
        strict: bool = False,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        ...

    def __init__(
        self,
        callback: _CallbackishT[_MessageCallbackSigT],
        name: str,
        /,
        *names: str,
        strict: bool = False,
        validate_arg_keys: bool = True,
        _wrapped_command: typing.Optional[tanjun.ExecutableCommand[typing.Any]] = None,
    ) -> None:
        """Initialise a message command group.

        Parameters
        ----------
        callback : tanjun.abc.MessageCallbackSig
            Callback to execute when the command is invoked.

            This should be an asynchronous callback which takes one positional
            argument of type [tanjun.abc.MessageContext][], returns [None][]
            and may use dependency injection to access other services.
        name
            The command name.
        *names
            Other names for the command.
        strict
            Whether this command group should only allow commands without spaces in their names.

            This allows for a more optimised command search pattern to be used and
            enforces that command names are unique to a single command within the group.
        validate_arg_keys
            Whether to validate that option keys match the command callback's signature.
        """
        super().__init__(callback, name, *names, validate_arg_keys=validate_arg_keys, _wrapped_command=_wrapped_command)
        self._commands = _internal.MessageCommandIndex(strict)

    def __repr__(self) -> str:
        return f"CommandGroup <{len(self._commands.commands)}: {self._names}>"

    @property
    def commands(self) -> collections.Collection[tanjun.MessageCommand[typing.Any]]:
        # <<inherited docstring from tanjun.abc.MessageCommandGroup>>.
        return self._commands.commands.copy()

    @property
    def is_strict(self) -> bool:
        return self._commands.is_strict

    def copy(self, *, parent: typing.Optional[tanjun.MessageCommandGroup[typing.Any]] = None) -> Self:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        inst = super().copy(parent=parent)
        inst._commands = self._commands.copy(parent=self)
        return inst

    def add_command(self, command: tanjun.MessageCommand[typing.Any], /) -> Self:
        """Add a command to this group.

        Parameters
        ----------
        command
            The command to add.

        Returns
        -------
        Self
            The group instance to enable chained calls.

        Raises
        ------
        ValueError
            If one of the command's names is already registered in a strict
            command group.
        """
        if self._commands.add(command):
            command.set_parent(self)

        return self

    def as_sub_command(
        self, name: str, /, *names: str, validate_arg_keys: bool = True
    ) -> collections.Callable[[_CallbackishT[_OtherCallbackSigT]], MessageCommand[_OtherCallbackSigT]]:
        """Build a message command in this group from a decorated callback.

        Parameters
        ----------
        name
            The command name.
        *names
            Other names for the command.
        validate_arg_keys
            Whether to validate that option keys match the command callback's signature.

        Returns
        -------
        collections.abc.Callable[[tanjun.abc.MessageCallbackSig], MessageCommand]
            The decorator callback used to make a sub-command.

            This can either wrap a raw command callback or another callable command instance
            (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][]).
        """

        def decorator(
            callback: typing.Union[_OtherCallbackSigT, _AnyCommandT[_OtherCallbackSigT]], /
        ) -> MessageCommand[_OtherCallbackSigT]:
            return self.with_command(as_message_command(name, *names, validate_arg_keys=validate_arg_keys)(callback))

        return decorator

    def as_sub_group(
        self, name: str, /, *names: str, strict: bool = False, validate_arg_keys: bool = True
    ) -> collections.Callable[[_CallbackishT[_OtherCallbackSigT]], MessageCommandGroup[_OtherCallbackSigT]]:
        """Build a message command group in this group from a decorated callback.

        Parameters
        ----------
        name
            The command name.
        *names
            Other names for the command.
        strict
            Whether this command group should only allow commands without spaces in their names.

            This allows for a more optimised command search pattern to be used and
            enforces that command names are unique to a single command within the group.
        validate_arg_keys
            Whether to validate that option keys match the command callback's signature.

        Returns
        -------
        collections.abc.Callable[[tanjun.abc.MessageCallbackSig], MessageCommand]
            The decorator callback used to make a sub-command group.

            This can either wrap a raw command callback or another callable command instance
            (e.g. [tanjun.MenuCommand][], [tanjun.MessageCommand][], [tanjun.SlashCommand][]).
        """

        def decorator(callback: _CallbackishT[_OtherCallbackSigT], /) -> MessageCommandGroup[_OtherCallbackSigT]:
            return self.with_command(
                as_message_command_group(name, *names, strict=strict, validate_arg_keys=validate_arg_keys)(callback)
            )

        return decorator

    def remove_command(self, command: tanjun.MessageCommand[typing.Any], /) -> Self:
        # <<inherited docstring from tanjun.abc.MessageCommandGroup>>.
        self._commands.remove(command)
        command.set_parent(None)
        return self

    def with_command(self, command: _AnyMessageCommandT, /) -> _AnyMessageCommandT:
        self.add_command(command)
        return command

    def bind_client(self, client: tanjun.Client, /) -> Self:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        super().bind_client(client)
        for command in self._commands.commands:
            command.bind_client(client)

        return self

    def bind_component(self, component: tanjun.Component, /) -> Self:
        # <<inherited docstring from tanjun.abc.ExecutableCommand>>.
        super().bind_component(component)
        for command in self._commands.commands:
            command.bind_component(component)

        return self

    def find_command(
        self, content: str, /, *, case_sensntive: bool = True
    ) -> collections.Iterable[tuple[str, tanjun.MessageCommand[typing.Any]]]:
        return self._commands.find(content, case_sensntive)

    async def execute(
        self,
        ctx: tanjun.MessageContext,
        /,
        *,
        hooks: typing.Optional[collections.MutableSet[tanjun.MessageHooks]] = None,
    ) -> None:
        # <<inherited docstring from tanjun.abc.MessageCommand>>.
        if ctx.message.content is None:
            raise ValueError("Cannot execute a command with a content-less message")

        if self._hooks:
            if hooks is None:
                hooks = set()

            hooks.add(self._hooks)

        case_sensitive = ctx.client.is_case_sensitive
        if ctx.component and ctx.component.is_case_sensitive is not None:
            case_sensitive = ctx.component.is_case_sensitive

        for name, command in self._commands.find(ctx.content, case_sensitive):
            if await command.check_context(ctx):
                content = ctx.content[len(name) :]
                ctx.set_triggering_name(ctx.triggering_name + " " + name)
                ctx.set_content(content.lstrip())
                await command.execute(ctx, hooks=hooks)
                return

        await super().execute(ctx, hooks=hooks)
